package agents;


import java.util.ArrayList;

import environment.Cell;
import event.TRNAEvent;

import objects.TRNAPropensity;
import objects.TranslationalRNA;

import utils.SequenceUtils;


/**
 * a class that stores abundances of tRNAs and information about various tRNA species
 * @author nrz2
 *
 */
public class TranslationalRNATable {
	//private Codon codons;
	private TranslationalRNA[] tRNAs;
	private int totalAbundance;
	private int chargedAbundance;
	private int busyAbundance;	
	private int unchargedAbundance;

	/**
	 * empty class constructor
	 */
	public TranslationalRNATable(double tRNAchargeRate){
		initialiseParameters(tRNAchargeRate);
	}

	/**
	 * local method used to initialise the parameters
	 */
	private void initialiseParameters(double tRNAchargeRate){
		tRNAs = new TranslationalRNA[SequenceUtils.tRNAtable.getNoOftRNAs()];
		for(int i=0;i<tRNAs.length; i++){
			tRNAs[i]=new TranslationalRNA(SequenceUtils.tRNAtable.gettRNA((byte)i));
			tRNAs[i].setRechargeRate(tRNAchargeRate);
		}
		this.totalAbundance = 0;
		for(TranslationalRNA tRNA:tRNAs){
			this.tRNAs[tRNA.getLabel()] = tRNA;
			this.totalAbundance+=tRNA.getAbundance();
		}
		this.chargedAbundance = totalAbundance;
		
		this.busyAbundance = 0;
		this.unchargedAbundance = 0;
	}
	
	
	
	
	
	
	
	
	
	
	
	/**
	 * returns the total number of tRNAs molecules
	 *
	 */
	public int getTotalAbundance(){
		return totalAbundance;
	}
	
	/**
	 * returns the total number of charged tRNAs
	 * @return
	 */
	public int getChargedAbundance(){
		return this.chargedAbundance;
	}
	
	/**
	 * returns the total number of busy tRNAs
	 * @return
	 */
	public int getBusyAbundance(){
		return this.busyAbundance;
	}
	
	
	/**
	 * returns the total number of uncharged tRNAs
	 * @return
	 */
	public int getUnchargedAbundance(){
		return this.unchargedAbundance;
	}
	
	/**
	 * computes the total abundance of the cognate tRNAs
	 * @param codonID
	 * @return
	 */
	public int getCognateAbundance(byte codonID){
		int result = 0;
		for(int i: SequenceUtils.tRNAtable.getCognates(codonID)){
			result+=tRNAs[i].getChargedAbundance();
		}
		return result;
	}
	
	/**
	 * computes the total abundance of the near cognate tRNAs
	 * @param codonID
	 * @return
	 */
	public int getNearCognateAbundance(byte codonID){
		int result = 0;
		for(int i: SequenceUtils.tRNAtable.getNearCognates(codonID)){
			result+=tRNAs[i].getChargedAbundance();
		}
		return result;
	}
	
	/**
	 * generates a string table with the tRNA population
	 */
	public String toString(){
		String result=tRNAs.length+"tRNAs\n";
		for(TranslationalRNA tRNA:tRNAs){
			result+=tRNA+"\n";
		}
		
		result+="\nCodons:\n";
		for(int i=0;i<SequenceUtils.codons.codon.length;i++){
			result+=SequenceUtils.codons.codon[i]+": (cognates) ";
			for(int j: SequenceUtils.tRNAtable.getCognates((byte)i)){
				result+=j+", ";
			}
			result+="; (near-cognates) ";
			for(int j: SequenceUtils.tRNAtable.getNearCognates((byte)i)){
				result+=j+", ";
			}
			result+="\n";
		}
		
		return result;
	}
	

	/**
	 * sets a charged tRNA to an busy one (bound to the ribosome)
	 * @param  tRNAID the id of the tRNA species that is affected
	 * @return true if the charged abundance is higher than 0 or false otherwise
	 */
	public boolean setBusy(int tRNAID){
		boolean result= tRNAs[tRNAID].setBusy();
		if(result){
			this.chargedAbundance--;
			this.busyAbundance++;	
		}
		return result;
	}
	
	
	/**
	 * sets a busy (bound to the ribosome) tRNA to an charged one 
	 * @param  tRNAID the id of the tRNA species that is affected
	 * @return true if the busy abundance is higher than 0 or false otherwise
	 */
	public boolean resetBusy(int tRNAID){
		boolean result=  tRNAs[tRNAID].resetBusy();
		if(result){
			this.busyAbundance--;
			this.chargedAbundance++;
                       
		}
		return result;
	}
	
	
	/**
	 * sets a busy (bound to the ribosome) tRNA to an uncharged one 
	 * @param  tRNAID the id of the tRNA species that is affected
	 * @return true if the busy abundance is higher than 0 or false otherwise
	 */
	public boolean uncharge(int tRNAID){
		boolean result=  tRNAs[tRNAID].uncharge();
		if(result){
			this.busyAbundance--;
			this.unchargedAbundance++;
		}
		return result;
	}
	
	
	/**
	 * sets an  uncharged  tRNA to an charged one
	 * @param  tRNAID the id of the tRNA species that is affected
	 * @return true if the uncharged abundance is higher than 0 or false otherwise
	 */
	public boolean charge(int tRNAID){
		boolean result=  tRNAs[tRNAID].charge();
		if(result){
			this.unchargedAbundance--;
			this.chargedAbundance++;
                     
		}
		return result;
	}
	
        
        
        
        
        
        
        
        
        
        
	public TRNAPropensity getChargedAbundances(byte codonID){
		TRNAPropensity result = new TRNAPropensity();
		//cognates
		ArrayList<Integer> cognatetRNAs = SequenceUtils.tRNAtable.getCognates(codonID);
		for(int i=0;i< cognatetRNAs.size();i++){
			result.propensitySum+=(double)tRNAs[cognatetRNAs.get(i)].getChargedAbundance();
			result.propensities.put(cognatetRNAs.get(i),new Double(tRNAs[cognatetRNAs.get(i)].getChargedAbundance()));
		}
		//near-cognates
		ArrayList<Integer> nearCognatetRNAs = SequenceUtils.tRNAtable.getNearCognates(codonID);
		for(int i=0;i< nearCognatetRNAs.size();i++){
			result.propensitySum+=(double)tRNAs[nearCognatetRNAs.get(i)].getChargedAbundance();
			result.propensities.put(nearCognatetRNAs.get(i),new Double(tRNAs[nearCognatetRNAs.get(i)].getChargedAbundance()));
		}
		return result;
	}
	
	
	/**
	 * computes the total abundance of the cognate tRNAs
	 * @param codonID
	 * @return
	 */
	public int[] getCognateChargedAbundance(byte codonID){
		ArrayList<Integer> cognatetRNAs = SequenceUtils.tRNAtable.getCognates(codonID);
		int[] result = new int[cognatetRNAs.size()];
		for(int i=0;i< cognatetRNAs.size();i++){
			result[i]=tRNAs[cognatetRNAs.get(i)].getChargedAbundance();
		}
		return result;
	}
	
	/**
	 * computes the total abundance of the near cognate tRNAs
	 * @param codonID nearCognatetRNAs
	 * @return
	 */
	public int[] getNearCognateChargedAbundance(byte codonID){
		ArrayList<Integer> nearCognatetRNAs = SequenceUtils.tRNAtable.getNearCognates(codonID);
		int[] result = new int[nearCognatetRNAs.size()];
		for(int i=0;i< nearCognatetRNAs.size();i++){
			result[i]=tRNAs[nearCognatetRNAs.get(i)].getChargedAbundance();
		}
		return result;
	}
	
	/**
	 * the table acts ... the tRNA gets charged
	 */
	public boolean act(TRNAEvent e, Cell cp){
		

		boolean result = charge(e.tRNAID);
		//cp.incrementTRNAchargedAvg(e.tRNAID, e.time);
		//cp.incrementTRNAunchargedAvg(e.tRNAID, e.time);
		cp.recordTRNAProportionsAvg(e.tRNAID, e.time);
		if(result){
			 
			//cp.updateElongationEventQueueBindingCodon(e.tRNAID,e.time);
                 (cp.getEventList()).gettRNABindingQ().addNexttRNAToRiboBindingEvent(cp.getCellTime(), cp.getRandomGenerator(), cp);    
                         
                         
		} else{
		
                    System.out.println("tRNA fail charging in TranslationRNATable:act: Probably a Gillespie problem");
                    System.exit(-1); 
                    
                    //cp.printDebugInfo("!!! not charged: "+e);
			//cp.scheduleNexttRNAEvent(e.tRNAID, e.time);
		}

		return result; 
	}
	
	/**
	 * returns the charged rate for a tRNA species
	 * @param tRNAID
	 * @return
	 */
	public double gettRNAchargeRate(int tRNAID){
		return this.tRNAs[tRNAID].getRechargeRate();
	}
	
	/**
	 * returns the charged tRNA abundance
	 * @param tRNAID the tRNA species
	 * @return
	 */
	public int gettRNAChargeAbundance(int tRNAID){
		return this.tRNAs[tRNAID].getChargedAbundance();
	}
	
	/**
	 * returns the busy tRNA abundance
	 * @param tRNAID the tRNA species
	 * @return
	 */
	public int gettRNABusyAbundance(int tRNAID){
		return this.tRNAs[tRNAID].getBusyAbundance();
	}
	
	/**
	 * returns the uncharged tRNA abundance
	 * @param tRNAID the tRNA species
	 * @return
	 */
	public int gettRNAunchargeAbundance(int tRNAID){
		return this.tRNAs[tRNAID].getUnchargedAbundance();
	}
	
	/**
	 * returns the total tRNA abundance
	 * @param tRNAID the tRNA species
	 * @return
	 */
	public int gettRNAabundance(int tRNAID){
		return this.tRNAs[tRNAID].getAbundance();
	}
}
