package environment;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.Random;

import event.Event;
import event.EventList;
import event.RibosomeEvent;
import event.TRNAEvent;

import agents.Ribosome;
import agents.TranslationalRNATable;

import objects.DynamicBehaviour;
import objects.FreedAgents;
import objects.MessengerRNA;
import objects.MessengerRNAType;
import objects.TRNAPropensity;
import utils.Constants;
import utils.FastaFileParser;
import utils.Gillespie;
import utils.SequenceUtils;
import utils.Utils;
import utils.mRNAcsvFile;

public class Cell {

	//parameters set externally
	private double stopTime;
	private double recordStartTime;
	private int noOfmRNAtype;
	private int noOfCopiesPermRNAAvg;
	private int mRNAcodonLengthMin;
	private int mRNAcodonLengthMax;
	private String mRNAfilename;
	private String mRNAabundanceFilename;
	private double IBSlengthAvg;
	private double IBSscanningRate;
	private double noOfCopiesPermRNAMin;
	private double noOfSynthesisedProteinsPermRNAmin;
	private int noOfRibosomes;
	private double ratioOfFreeRibosomes;
	private String resultsFilepath;
	private String resultsFilename;
	private String debugOutputFile;
	private boolean debugMode;
	private boolean approximateSystem;
	private boolean unbindsFromIBSifFree;
	private boolean unbindsFromterminationIfFree;
	private boolean stochasticmRNACopyNumber;
	private boolean printPolyosomeSize;
	private boolean unbindsFromCodonIfBusy;
	private int collisionsBeforeUnbinds;
	private int polyosomeSizePrintPoints;
	private boolean nonGognateIntroduceWaitingTime; 
	private double nonGognateWaitingTime; 
	private double nearGognateWaitingTime; 
	private boolean nonCognateChangeBindingRate;
	public static  String tRNAAbundaceFile;
	private double tRNAbindingRate;
	private double tRNAchargeRate;
	private double ribosomeAffinityAvg;
	private double ribosomeAffinityStd;
	private double IBSmoveRateMax;
	private double IBSmoveRateMin;
	private int noOfRibosomeSitesBefore;
	private int noOfRibosomeSitesAfter;
	private int noOfRibosomeSitesTotal;
	private int noOfTranslationPerRibosomemRNAMax;
	private int noOfTranslationPerRibosomemRNAMin;
	private double elongationCognateProbability;
	private double elongationNearCognateProbability;
	private double elongationNearCognateProbabilityExitRecognition;
	private double elongationCognateProbabilityExitRecognition;
	private double elongationCognateRate;
	private double elongationNearCognateRate;
	public static double elongationTranslocationRate;
	private double elongationNonCognateProbability;
	private long randommRNASeed;
	private boolean elongationBindingEventsUseGillespeDirect;
	private boolean stallEventsWait;
	
	//internal parameters
        private int[]  mRNABoundList;
	private MessengerRNAType[] mRNAtype; 
	public MessengerRNA[] mRNA;
	private Ribosome[] ribosomes;
	private ArrayList<Integer> freeRibosomes;
	private PriorityQueue<Integer> occupiedRibosomes;
	private ArrayList<ArrayList<Integer>> freemRNAs;
	private double cellTime;
	public TranslationalRNATable tRNATabe;
	private double nonCognateUnbindingRate;
	private double nonCognateMMConstant;
	private int numberOfFreeBoundRibosomes;
	private int totalmRNACopyNumber;
	private Random randomGenerator;
	private EventList eventQueue; 
	
	
	// output parameters
	private int[] mRNATranslated;
	//private double[] productionRate;
	private long succesfullElongations;
	private long unsuccesfullElongations;
	private int polyosomeSize;
	private ArrayList<DynamicBehaviour<Integer>> ps;//polyosomeSize vector
	private long[] cognateElongated; 
	private long[] nearCognateElongated; 
	private double[] cognateElongationTimeTotal; 
	private double[] nearCognateElongationTimeTotal; 
	private long[] codonCount; 
	private long codonCountTotal;
	private long[] cognateAttempts; 
	private long[] nearCognateAttempts; 
	private long[] cognateAttemptsTotal; 
	private long[] nearCognateAttemptsTotal; 
	private int[] ribosomesPerMRNA;
	private double[] ribosomesPerMRNAAvg;
	private int[] ribosomesPerMRNACount;
	private double[] ribosomesPerMRNALast;
	private long unsuccesfullEventRandomWalkCount;
	private long unsuccesfullEventTranslocationCount;
	private long succesfullEventRandomWalkCount;
	private long succesfullEventTranslocationCount;
	private File parameterFile;
	
	private double[] tRNAchargedAvg; 
	private int[] tRNAchargedCount; 
	private double[] tRNAchargedLast; 
	private double[] tRNAbusyAvg; 
	private int[] tRNAbusyCount; 
	private double[] tRNAbusyLast; 
	private double[] tRNAunchargedAvg; 
	private int[] tRNAunchargedCount; 
	private double[] tRNAunchargedLast; 
	
	//debug params
	private double simulationTime;
	private int simulationTimeReadings;
	
	/**
	 * constructor that initialises a cell with mRNAs
	 */
	public Cell(String parametersFile){
		//initialise the internal parameters
		initialiseInputParameters();
		
		//load input parameters from file
		loadParametersFromFile(parametersFile);
		
		//
		generateParameterFile(parametersFile);
		printDebugInfo("parameters printed to file: "+this.resultsFilepath+this.parameterFile.getName(),true);
		
		//check input data consistency
		checkInputDataConsistency();
		
		//initialise internal parameters
		initialiseInternalParameters();
		
		//initialise output parameters
		initialiseOutputParameters();
		printDebugInfo("========================================",true);
		
	}
	
	/**
	 * generates the mRNA types randomly
	 */
	private void generateRandommRNA(){
		this.mRNAtype = new MessengerRNAType[this.noOfmRNAtype];
		NumberFormat formatter = new DecimalFormat("0000");
		String name = "";
		this.totalmRNACopyNumber = 0;
		for(int i=0;i<this.noOfmRNAtype;i++){
			name = "mRNA"+formatter.format(i);
			this.mRNAtype[i] = new MessengerRNAType(name,this.randomGenerator, this.mRNAcodonLengthMin, this.mRNAcodonLengthMax, this.ribosomeAffinityAvg, this.IBSmoveRateMin, this.IBSmoveRateMax);
			this.mRNAtype[i].setCopyNumber(computemRNACopyNo());
			this.mRNAtype[i].setSSabundace(this.mRNAtype[i].getCopyNumber());
			this.totalmRNACopyNumber+=this.mRNAtype[i].getCopyNumber();
		}
	}
	
	
	
	/**
	 * computes the copy number for a specific mRNA
	 * @return
	 */
	private int computemRNACopyNo(){
		int buffer;
		if(this.stochasticmRNACopyNumber){
			buffer = ((int)(this.noOfCopiesPermRNAAvg/2))+this.randomGenerator.nextInt(this.noOfCopiesPermRNAAvg+1);
			if(buffer<0){
				buffer=0;
			}
		} else{
			buffer=noOfCopiesPermRNAAvg;
		}
		return buffer;
	}
	
	/**
	 * loads the mRNAs from a file
	 * @param filename the file where to load the mRNAs from
	 * @return true if the file exists or false otherwise
	 */
	private boolean loadmRNAFromFile(String filename){
		boolean result=false;
		File f=new File(filename);
		this.totalmRNACopyNumber = 0;
		if(f.exists()){
			ArrayList<MessengerRNAType> mRNAs = FastaFileParser.fileParser(filename);
			if(!mRNAs.isEmpty()){
				this.mRNAtype = new  MessengerRNAType[mRNAs.size()];
				for(int i=0;i<mRNAs.size();i++){
					this.mRNAtype[i] = mRNAs.get(i);
					this.totalmRNACopyNumber+=this.mRNAtype[i].getCopyNumber();
				}
				this.noOfmRNAtype = mRNAs.size();
				result=true;
				
				
			}
		}
		return result;
	}
	
	/**
	 * loads mRNA data from file
	 * @param filename
	 * @return
	 */
	private boolean loadmRNADataFromFile(String filename){
		boolean result=false;
		File f=new File(filename);
		int index;
		double minLength = this.IBSmoveRateMin, maxLength = this.IBSmoveRateMax;

		if(f.exists()){
			ArrayList<MessengerRNAType> mRNAs = mRNAcsvFile.fileParser(filename, this.IBSlengthAvg, this.noOfCopiesPermRNAMin, this.noOfSynthesisedProteinsPermRNAmin);
			minLength = Integer.MAX_VALUE; 
			maxLength = Integer.MIN_VALUE;
			if(!mRNAs.isEmpty()){
				result=true;
			}
			for(MessengerRNAType mt: mRNAs){
				index = this.getmRNAtypeID(mt.getName());
				if(index>=0){
                                    
                                    
                                    //THIS NEEDS TO BE CHANGED NOW
					this.mRNAtype[index].setIBSlength(mt.getIBSlength());
					this.mRNAtype[index].setSSabundace(mt.getSSabundace());
					this.mRNAtype[index].setProteinSynthesisRate(mt.getProteinSynthesisRate());

					//min IBS length
					if(minLength > mt.getIBSlength()){
						minLength = mt.getIBSlength();
					}
					
					//max IBS length
					if(maxLength < mt.getIBSlength()){
						maxLength = mt.getIBSlength();
					}
				} else{
					//System.out.println(mt.getName()+" could not be found");
				}
			}
		}
		double cycles = (this.noOfTranslationPerRibosomemRNAMax+this.noOfTranslationPerRibosomemRNAMin)/2;
		double ribosomes = this.noOfRibosomes*this.ratioOfFreeRibosomes;
		double buffer = 0;
		this.totalmRNACopyNumber = 0;
		for(int i=0;i<this.mRNAtype.length;i++){
			//IBS move rate 
			if(this.mRNAtype[i].getIBSlength() <= 0){
				this.mRNAtype[i].setIBSlength(this.IBSlengthAvg);
			}
//COMPUTE IBS RANDOM WALK RATE HERE
			//buffer = this.IBSmoveRateMin+((this.mRNAtype[i].getIBSlength()-minLength)/(maxLength-minLength))*(this.IBSmoveRateMax-this.IBSmoveRateMin);
			//buffer = this.IBSscanningRate/this.mRNAtype[i].getIBSlength();
			//buffer = this.IBSscanningRate;//Changed by CHU
			buffer = this.IBSscanningRate;//Changed by CHU
                        this.mRNAtype[i].setIBSmoveRate(buffer);
			
			//copy number
			if(this.mRNAtype[i].getSSabundace()<0){
				this.mRNAtype[i].setSSabundace(this.noOfCopiesPermRNAMin);
			}
			if(this.mRNAtype[i].getSSabundace() > 1){
				this.mRNAtype[i].setCopyNumber((int)Math.round(this.mRNAtype[i].getSSabundace()));
			} else if(this.mRNAtype[i].getSSabundace()>0){
				if(this.randomGenerator.nextDouble() < this.mRNAtype[i].getSSabundace()){
					this.mRNAtype[i].setCopyNumber(1);
				}
			} else{
				this.mRNAtype[i].setCopyNumber(0);
			}
			this.totalmRNACopyNumber+=this.mRNAtype[i].getCopyNumber();
			
			//ribosome affinity
			if(this.mRNAtype[i].getProteinSynthesisRate() <0 ){
				this.mRNAtype[i].setProteinSynthesisRate(0);
			}
			this.mRNAtype[i].setRibosomeAffinity(computeRibosomeAffinity(this.mRNAtype[i].getProteinSynthesisRate(), cycles,ribosomes,this.mRNAtype[i].getCopyNumber()));
		}
		
		//removes the mRNAs that are never transcribed
		//System.out.println(this.mRNAtype.length+" mRNAs");
		cleanmRNAtypes();
	//	System.out.println(this.mRNAtype.length+" mRNAs");
		
		return result;
	}
	
	/**
	 * removes the mRNAtypes that are never synthesised
	 */
	public void cleanmRNAtypes(){
		ArrayList<MessengerRNAType> mRNAtypes = new ArrayList<MessengerRNAType>();
		//get the non zero elements
		for(int i=0;i<this.mRNAtype.length;i++){
			if(this.mRNAtype[i].getCopyNumber() > 0 && this.mRNAtype[i].getRibosomeAffinity() > 0){
				mRNAtypes.add(this.mRNAtype[i]);
			}
		}
		
		//put back only the non-zero elements
		this.totalmRNACopyNumber = 0;
		this.noOfmRNAtype = mRNAtypes.size();
		this.mRNAtype = new MessengerRNAType[mRNAtypes.size()];
		for(int i=0;i<mRNAtypes.size();i++){
			this.mRNAtype[i] = mRNAtypes.get(i);
			this.totalmRNACopyNumber += this.mRNAtype[i].getCopyNumber();
		}
		
	}
	
	/**
	 * computes the affinity of the ribosome for a mRNA stranf
	 * @param proteinSynthesisRate the number of molecules/min in a single cell
	 * @return 
	 */
	private double computeRibosomeAffinity(double proteinSynthesisRate, double cycles, double ribosomes, double copyNumber){
		//return 60*proteinSynthesisRate/(500000-6.25*proteinSynthesisRate);
		//return proteinSynthesisRate/(60*cycles*ribosomes*copyNumber);
		return proteinSynthesisRate;//changed by Chu This is really the rate constant of the ribosome for mRNA
		 
	}
	
        
            public   EventList getEventList(){return eventQueue;}
        
        
	private void initialiseDebugParameters(){
		this.simulationTime = 0;
		this.simulationTimeReadings = 0;
	}
	
	/**
	 * initialises the internal parameters
	 */
	private void initialiseInternalParameters(){

		//create the random number generator
		randomGenerator = new Random();
		if(this.randommRNASeed>0){
			this.printDebugInfo("random numbers use a common seed");
			randomGenerator = new Random(this.randommRNASeed);
		} else{
			this.printDebugInfo("random numbers use a different seed");
		}
		
		//create the ribosomes
		this.freeRibosomes = new ArrayList<Integer>();
		ribosomes = new Ribosome[this.noOfRibosomes];
		
                
                
                for(int i=0; i<this.noOfRibosomes;i++){
			this.ribosomes[i] = new Ribosome(i);
			this.freeRibosomes.add(i);
		}
		 
		
		//System.out.println(this.mRNAfilename); 
		//create the mRNA species
		if(!loadmRNAFromFile(this.mRNAfilename)){
			generateRandommRNA();
			printDebugInfo(this.mRNAtype.length + " mRNA species were randomly generated");
		} else{
			loadmRNADataFromFile(this.mRNAabundanceFilename);
			printDebugInfo(this.mRNAtype.length + " mRNA species imported from file "+this.mRNAfilename);
		}
		
		//create the mRNAs 
		this.freemRNAs=new ArrayList<ArrayList<Integer>>();
		this.occupiedRibosomes =  new PriorityQueue<Integer>();

		int index=0;
		ArrayList<Integer> bufferList;
		this.mRNA = new MessengerRNA[this.totalmRNACopyNumber];
		for(int i=0;i<this.noOfmRNAtype;i++){
			bufferList = new ArrayList<Integer>();
			for(int j=0;j<this.mRNAtype[i].getCopyNumber();j++){
				bufferList.add(index);
				this.mRNA[index] = new MessengerRNA(this.mRNAtype[i].getName(),i,this.mRNAtype[i].getLength());
				index++;
			}
			this.freemRNAs.add(bufferList);
		}			
		
		printDebugInfo(this.totalmRNACopyNumber + " mRNA molecules were created");

//for(int i=0; i<mRNA.length; i++) System.out.println(i + "   " + mRNA[i].getName()   ); //CHUCORR
		

		this.cellTime=0;
		
		//initialise the tRNA table
		this.tRNATabe = new TranslationalRNATable(this.tRNAchargeRate);
		//this.tRNATabe.loadFile(this.tRNAAbundaceFile,this);
		printDebugInfo("tRNA data was loaded from file " + this.tRNAAbundaceFile);
		
		//compute the Michaelis menten constant
		this.nonCognateMMConstant = (double) (1/this.nonGognateWaitingTime)/this.tRNAbindingRate;
		this.numberOfFreeBoundRibosomes = 0;
		
		eventQueue = new EventList(this);
		
		// reset the random number generator
		randomGenerator = new Random();
		if(this.randommRNASeed>0){
			randomGenerator = new Random(this.randommRNASeed);
		}
	}
	

	/**
	 * 
	 */
	public double getCellTime(){ return cellTime; }	
	



	/**
	 * initialises the output parameters
	 */
	private void initialiseOutputParameters(){
		//initialise the number of translated mRNAs
		mRNATranslated = new int[noOfmRNAtype];
		//productionRate = new double[noOfmRNAtype];
		ribosomesPerMRNA = new int[noOfmRNAtype];
		ribosomesPerMRNAAvg = new double[noOfmRNAtype];
		ribosomesPerMRNACount =  new int[noOfmRNAtype];
		ribosomesPerMRNALast = new double[noOfmRNAtype];
		for(int i=0;i<this.noOfmRNAtype;i++){
			mRNATranslated[i] = 0;
			//productionRate[i] = 0;
			ribosomesPerMRNA[i] = 0;
			ribosomesPerMRNAAvg[i] = 0;
			ribosomesPerMRNACount[i] = 0;
			ribosomesPerMRNALast[i] = this.recordStartTime;
		}
		
		//polyosome size
		polyosomeSize = 0;
		ps = new ArrayList<DynamicBehaviour<Integer>>();
		ps.add(new DynamicBehaviour<Integer>(0,0));
		
	
		//number of elongations
		this.succesfullElongations = 0;
		this.unsuccesfullElongations = 0;
		
		//speed and accuracy of elongation per codon
		int noOfCodons = SequenceUtils.codons.codon.length;
		this.cognateElongated = new long[noOfCodons]; 
		this.nearCognateElongated = new long[noOfCodons]; 
		this.cognateElongationTimeTotal = new double[noOfCodons]; 
		this.nearCognateElongationTimeTotal = new double[noOfCodons];
		this.codonCount = new long[noOfCodons]; 
		this.cognateAttempts =  new long[noOfCodons]; 
		this.nearCognateAttempts =  new long[noOfCodons]; 
		this.cognateAttemptsTotal =  new long[noOfCodons]; 
		this.nearCognateAttemptsTotal =  new long[noOfCodons]; 
		for(int i=0;i<noOfCodons;i++){
			this.cognateElongated[i] = 0; 
			this.nearCognateElongated[i] = 0; 
			this.cognateElongationTimeTotal[i] = 0; 
			this.nearCognateElongationTimeTotal[i] = 0; 
			this.codonCount[i] = 0;
			this.cognateAttempts[i] =0;
			this.nearCognateAttempts[i] =0;
			this.cognateAttemptsTotal[i] =0;
			this.nearCognateAttemptsTotal[i] =0;
		}
		
		this.unsuccesfullEventRandomWalkCount=0;
		this.unsuccesfullEventTranslocationCount=0;
		this.succesfullEventRandomWalkCount=0;
		this.succesfullEventTranslocationCount=0;
                 
                this.mRNABoundList = new int[totalmRNACopyNumber];		//CHU  
for(int i=0; i< totalmRNACopyNumber; i++) mRNABoundList[i]=0; //CHU
		this.tRNAchargedAvg = new double[SequenceUtils.tRNAtable.getNoOftRNAs()]; 
		this.tRNAchargedCount = new int[SequenceUtils.tRNAtable.getNoOftRNAs()]; 
		this.tRNAchargedLast = new double[SequenceUtils.tRNAtable.getNoOftRNAs()]; 
		this.tRNAbusyAvg = new double[SequenceUtils.tRNAtable.getNoOftRNAs()]; 
		this.tRNAbusyCount = new int[SequenceUtils.tRNAtable.getNoOftRNAs()]; 
		this.tRNAbusyLast = new double[SequenceUtils.tRNAtable.getNoOftRNAs()]; 
		this.tRNAunchargedAvg = new double[SequenceUtils.tRNAtable.getNoOftRNAs()]; 
		this.tRNAunchargedCount = new int[SequenceUtils.tRNAtable.getNoOftRNAs()]; 
		this.tRNAunchargedLast = new double[SequenceUtils.tRNAtable.getNoOftRNAs()]; 
		for(int i=0; i<SequenceUtils.tRNAtable.getNoOftRNAs();i++){
			this.tRNAchargedAvg[i] = 0;
			this.tRNAchargedCount[i] = 0;
			this.tRNAchargedLast[i] = this.recordStartTime;
			this.tRNAbusyAvg[i] = 0;
			this.tRNAbusyCount[i] = 0;
			this.tRNAbusyLast[i] = this.recordStartTime;
			this.tRNAunchargedAvg[i] = 0;
			this.tRNAunchargedCount[i] = 0;
			this.tRNAunchargedLast[i] = this.recordStartTime;
		}
		
	}

	/**CHU ADDED THIS
	 * Records how many ribosomes are bound to mRNA
	 * @return
	 */
	public void addRibosomeTo( int mRNACode){
		mRNABoundList[mRNACode]++;
	}
	


	/**CHU ADDED THIS
	 * Records how many ribosomes are bound to mRNA
	 * @return
	 */
	public void removeRibosomeFrom(int mRNACode){
		mRNABoundList[mRNACode]--;
     
if(mRNABoundList[mRNACode]< 0) 
    
{System.out.println("This is not right at all!!!!");}
	}
	


	/**CHU ADDED THIS
	 * Records how many ribosomes are bound to mRNA
	 * @return
	 */
	public int reportNumberOfRibsomesOn(int mRNACode){
	return	(mRNABoundList[mRNACode]);
	}
	





	/**
	 * checks if if the ribosome cannot move then it is put in a waiting list 
	 * @return
	 */
	public boolean doesStallEventsWait(){
		return this.stallEventsWait;
	}
	
	/**
	 * returns the probability that a bound near-cognate will exit from codon recognition	
	 * @return
	 */
	public double getElongationNearCognateProbabilityExitRecognition(){
		return this.elongationNearCognateProbabilityExitRecognition;
	}

	/**
	 * returns the number of consecutive collisions before the ribosome unbinds
	 * @return
	 */
	public int getCollisionsBeforeUnbinds(){
		return this.collisionsBeforeUnbinds;
	}
	
	
	/**
	 * returns the probability that a bound cognate will exit from codon recognition	
	 * @return
	 */
	public double getElongationCognateProbabilityExitRecognition(){
		return this.elongationCognateProbabilityExitRecognition;
	}
	/**
	 * increments the number of ribosome per mRNA
	 * @param mRNA the ID of the mRNA
	 */
	public void incrementRibosomesPerMRNA(int mRNAID, double time){
		int mRNAtype = this.mRNA[mRNAID].getType();
		this.ribosomesPerMRNA[mRNAtype]++;
		incrementRibosomesPerMRNAAvg(mRNAtype,time);
		
	}
	
	/**
	 * increments the number of ribosome per mRNA
	 * @param mRNA the ID of the mRNA
	 */
	public void decrementRibosomesPerMRNA(int mRNAID, double time){
		int mRNAtype = this.mRNA[mRNAID].getType();
		this.ribosomesPerMRNA[mRNAtype]--;
		incrementRibosomesPerMRNAAvg(mRNAtype,time);
	}
	
	/**
	 * recomputes the avg for the polyosome size
	 * @param mRNAtype
	 * @param time
	 */
	private void incrementRibosomesPerMRNAAvg(int mRNAtype, double time){
		if(time > this.recordStartTime){
			//this.ribosomesPerMRNAAvg[mRNAtype]=this.ribosomesPerMRNAAvg[mRNAtype] - (double) (this.ribosomesPerMRNAAvg[mRNAtype]-this.ribosomesPerMRNA[mRNAtype])/(this.ribosomesPerMRNACount[mRNAtype]+1);
			this.ribosomesPerMRNAAvg[mRNAtype] = this.ribosomesPerMRNAAvg[mRNAtype]- (double) (this.ribosomesPerMRNAAvg[mRNAtype]-this.ribosomesPerMRNA[mRNAtype])*(time-this.ribosomesPerMRNALast[mRNAtype])/(time);
			this.ribosomesPerMRNALast[mRNAtype] = time;
			this.ribosomesPerMRNACount[mRNAtype]++;
		}
	}
	
	/**
	 * increments the total number of charged tRNAs
	 * @param tRNAID
	 * @param time
	 */
	/*public void incrementTRNAchargedAvg(int tRNAID, double time){
		if(time > this.recordStartTime){
			//this.tRNAchargedAvg[tRNAID]=(double)(this.tRNAchargedAvg[tRNAID]*this.tRNAchargedCount[tRNAID]+this.tRNATabe.gettRNAChargeAbundance(tRNAID))/(this.tRNAchargedCount[tRNAID]+1);
			//this.tRNAchargedAvg[tRNAID]=this.tRNAchargedAvg[tRNAID] - (double) (this.tRNAchargedAvg[tRNAID]-this.tRNATabe.gettRNAChargeAbundance(tRNAID))/(this.tRNAchargedCount[tRNAID]+1);
			this.tRNAchargedAvg[tRNAID]=this.tRNAchargedAvg[tRNAID] - (double) (this.tRNAchargedAvg[tRNAID]-this.tRNATabe.gettRNAChargeAbundance(tRNAID))*(time-this.tRNAchargedLast[tRNAID])/(time);
			this.tRNAchargedLast[tRNAID] = time;
			this.tRNAchargedCount[tRNAID]++;
		}
	}*/
	
	
	/**
	 * increments the total number of uncharged tRNAs
	 * @param tRNAID
	 * @param time
	 */
	/*public void incrementTRNAunchargedAvg(int tRNAID, double time){
		if(time > this.recordStartTime){
			//this.tRNAunchargedAvg[tRNAID]=(double)(this.tRNAunchargedAvg[tRNAID]*this.tRNAunchargedCount[tRNAID]+this.tRNATabe.gettRNAunchargeAbundance(tRNAID))/(this.tRNAunchargedCount[tRNAID]+1);
			//this.tRNAunchargedAvg[tRNAID]=this.tRNAunchargedAvg[tRNAID] - (double)(this.tRNAunchargedAvg[tRNAID]-this.tRNATabe.gettRNAunchargeAbundance(tRNAID))/(this.tRNAunchargedCount[tRNAID]+1);
			this.tRNAunchargedAvg[tRNAID]=this.tRNAunchargedAvg[tRNAID] - (double) (this.tRNAunchargedAvg[tRNAID]-this.tRNATabe.gettRNAunchargeAbundance(tRNAID))*(time-this.tRNAunchargedLast[tRNAID])/(time);
			this.tRNAunchargedLast[tRNAID] = time;
			this.tRNAunchargedCount[tRNAID]++;
		}
	}*/
	
	

	
	/**
	 * increments the total number of busy tRNAs
	 * @param tRNAID
	 * @param time
	 */
	/*public void incrementTRNAbusyTotal(int tRNAID, double time){
		if(time > this.recordStartTime){
			//this.tRNAbusyAvg[tRNAID]=(double)(this.tRNAbusyAvg[tRNAID]*this.tRNAbusyCount[tRNAID]+this.tRNATabe.gettRNABusyAbundance(tRNAID))/(this.tRNAbusyCount[tRNAID]+1);
			//tthis.tRNATabe.gettRNAunchargeAbundance(tRNAID)his.tRNAbusyAvg[tRNAID]=this.tRNAbusyAvg[tRNAID] - (double)(this.tRNAbusyAvg[tRNAID] -this.tRNATabe.gettRNABusyAbundance(tRNAID))/(this.tRNAbusyCount[tRNAID]+1);
			this.tRNAbusyAvg[tRNAID]=this.tRNAbusyAvg[tRNAID] - (double) (this.tRNAbusyAvg[tRNAID]-this.tRNATabe.gettRNABusyAbundance(tRNAID))*(time-this.tRNAbusyLast[tRNAID])/(time);
			this.tRNAbusyLast[tRNAID] = time;
			this.tRNAbusyCount[tRNAID]++;
		}
	}*/
	
	/**
	 * records the new average for the proportions of tRNAs (charged/busy/uncharged)
	 * @param tRNAID
	 * @param time
	 */
	public void recordTRNAProportionsAvg(int tRNAID, double time){
		if(time > this.recordStartTime){
			//this.tRNAchargedAvg[tRNAID]=(double)(this.tRNAchargedAvg[tRNAID]*this.tRNAchargedCount[tRNAID]+this.tRNATabe.gettRNAChargeAbundance(tRNAID))/(this.tRNAchargedCount[tRNAID]+1);
			//this.tRNAchargedAvg[tRNAID]=this.tRNAchargedAvg[tRNAID] - (double) (this.tRNAchargedAvg[tRNAID]-this.tRNATabe.gettRNAChargeAbundance(tRNAID))/(this.tRNAchargedCount[tRNAID]+1);
			int totalAbundance = this.tRNATabe.gettRNAabundance(tRNAID);
			int chargedAbundance = this.tRNATabe.gettRNAChargeAbundance(tRNAID);
			int busyAbundance = this.tRNATabe.gettRNABusyAbundance(tRNAID);
			int unchargedAbundance = this.tRNATabe.gettRNAunchargeAbundance(tRNAID);
			if(chargedAbundance +busyAbundance+unchargedAbundance  != totalAbundance){
				this.printDebugInfo("error: number of tRNAs of type t"+tRNAID+" do not match (" + totalAbundance +" <> "+(chargedAbundance +busyAbundance+unchargedAbundance)+")at time "+time);
			} else{
				this.tRNAchargedAvg[tRNAID]=this.tRNAchargedAvg[tRNAID] - (double) (this.tRNAchargedAvg[tRNAID]-((double)chargedAbundance/totalAbundance))*(time-this.tRNAchargedLast[tRNAID])/(time);
				this.tRNAbusyAvg[tRNAID]=this.tRNAbusyAvg[tRNAID] - (double) (this.tRNAbusyAvg[tRNAID]-((double)busyAbundance/totalAbundance))*(time-this.tRNAbusyLast[tRNAID])/(time);
				this.tRNAunchargedAvg[tRNAID]=this.tRNAunchargedAvg[tRNAID] - (double) (this.tRNAunchargedAvg[tRNAID]-((double)unchargedAbundance/totalAbundance))*(time-this.tRNAunchargedLast[tRNAID])/(time);
			} 
			
			
			
			
			this.tRNAchargedLast[tRNAID] = time;
			this.tRNAbusyLast[tRNAID] = time;
			this.tRNAunchargedLast[tRNAID] = time;

			this.tRNAchargedCount[tRNAID]++;
			this.tRNAbusyCount[tRNAID]++;
			this.tRNAunchargedCount[tRNAID]++;
		}
	}
	
	/**
	 * increments unsuccesfullEventRandomWalkCount
	 */
	public void incrementUnsuccesfullEventRandomWalkCount(double time){
		if(time > this.recordStartTime){
			this.unsuccesfullEventRandomWalkCount++;
		}
	}
	
	
	/**
	 * increments unsuccesfullEventRandomWalkCount
	 */
	public void incrementUnsuccesfullTranslocationCount(double time){
		if(time > this.recordStartTime){
			this.unsuccesfullEventTranslocationCount++;
		}
	}
	
	
	
	/**
	 * increments unsuccesfullEventRandomWalkCount
	 */
	public void incrementSuccesfullEventRandomWalkCount(double time){
		if(time > this.recordStartTime){
			this.succesfullEventRandomWalkCount++;
		}
	}
	
	
	/**
	 * increments unsuccesfullEventRandomWalkCount
	 */
	public void incrementSuccesfullTranslocationCount(double time){
		if(time > this.recordStartTime){
			this.succesfullEventTranslocationCount++;
		}
	}
	
	/**
	 * increments the number of cognate elongations of a codon
	 * @param codonID
	 */
	public void incrementCognateElongated(int codonID, double time){
		if(time > this.recordStartTime){
			cognateElongated[codonID]++;
		}
	}
	
	
	/**
	 * increments the number of cognate binding attempts of a codon
	 * @param codonID
	 */
	public void incrementCognateAttempts(int codonID, double time){
		if(time > this.recordStartTime){
			this.cognateAttempts[codonID]++;
		}
	}
	
	/**
	 * increments the number of near-cognate binding attempts of a codon
	 * @param codonID
	 */
	public void incrementNearCognateAttempts(int codonID, double time){
		if(time > this.recordStartTime){
			this.nearCognateAttempts[codonID]++;
		}
	}
	
	/**
	 * increments the number of near-cognate elongations of a codon
	 * @param codonID
	 */
	public void incrementNearCognateElongated(int codonID, double time){
		if(time > this.recordStartTime){
			nearCognateElongated[codonID]++;
		}
	}
	
	/**
	 * adds a time to the total elongation time of a specific codon
	 * @param codonID
	 * @param time
	 */
	public void addCognateElongationTimeTotal(int codonID, double addTime, double currentTime){
		if(currentTime > this.recordStartTime){
			cognateElongationTimeTotal[codonID]+=addTime;
		}
	}
	
	/**
	 * adds a time to the total elongation time of a specific codon
	 * @param codonID
	 * @param time
	 */
	public void addNearCognateElongationTimeTotal(int codonID, double addTime, double currentTime){
		if(currentTime > this.recordStartTime){
			nearCognateElongationTimeTotal[codonID]+=addTime;
		}
	}
	
	
	/**
	 * adds a time to the total cognat attempts time of a specific codon
	 * @param codonID
	 * @param time
	 */
	public void addCognateAttemptsTotal(int codonID, int attempts, double time){
		if(time > this.recordStartTime){
			this.cognateAttemptsTotal[codonID]+=attempts;
		}
	}	
	
	/**
	 * adds a time to the total near-cognat attempts time of a specific codon
	 * @param codonID
	 * @param time
	 */
	public void addNearCognateAttemptsTotal(int codonID, int attempts, double time){
		if(time > this.recordStartTime){
			this.nearCognateAttemptsTotal[codonID]+=attempts;
		}
	}	
	
	/**
	 * returns the ID of the mRNA type providing the name
	 * @param name
	 * @return
	 */
	public int getmRNAtypeID(String name){
		int result = -1;
		for(int i = 0;i < this.mRNAtype.length;i++){
			if(this.mRNAtype[i].getName().equals(name)){
				result =i;
				break;
			}
		}
		return result;
	}
	
	/**
	 * returns the affinity of the ribosome for a specific mRNA
	 * @param mRNAID
	 * @return
	 */
	public double getmRNARibosomeAffinity(int mRNAtype){
		return this.mRNAtype[mRNAtype].getRibosomeAffinity();
	}
	
	/**
	 * returns the copynumber of a mRNA
	 * @param mRNAID
	 * @return
	 */
	public int getmRNACopyNumber(int mRNAID){
		return this.mRNAtype[mRNAID].getCopyNumber();
	}
	
	/**
	 * gets the number of free mRNA for a specific type
	 * @param mRNAtype
	 * @return
	 */
	public int getFreemRNACopyNumber(int mRNAtype){
		return freemRNAs.get(mRNAtype).size();
	}
	
        
        
        public int getFreemRNACopyNumber(){
	
            int k=0;
            for(int i=0;i<freemRNAs.size();i++){
            
             k+= freemRNAs.get(i).size();
                
            }
                
            return k;
	}
        
	
	/**
	 * returns no of mRNA species
	 * @return
	 */
	public int getNoOfmRNATypes(){
		return this.noOfmRNAtype;
	}
	
	/**
	 * returns the maximum number of translations of a Ribosome per mRNA
	 * @return
	 */
	public int getNoOfTranslationsPerRibosomemRNAMax(){
		return this.noOfTranslationPerRibosomemRNAMax;
	}
	
	/**
	 * returns the number of occupied ribosome sites before
	 * @return
	 */
	public int getNoOfRibosomeSitesBefore(){
		return this.noOfRibosomeSitesBefore;
	}
	
	public double gettRNAchargeRate(){
		return this.tRNAchargeRate;
	}
	
	/**
	 * returns the number occupied sites after
	 * @return
	 */
	public int getNoOfRibosomeSitesAfter(){
		return this.noOfRibosomeSitesAfter;
	}
	/**
	 * returns the minimum number of translations of a Ribosome per mRNA
	 * @return
	 */
	public int getNoOfTranslationsPerRibosomemRNAMin(){
		return this.noOfTranslationPerRibosomemRNAMin;
	}
	
	/**
	 * returns the random number generator
	 * @return
	 */
	public Random getRandomGenerator(){
		return this.randomGenerator;
	}
	
	/**
	 * increments the polyosome size
	 *
	 */
	public void incrementPolyosomeSize(double time){
		this.polyosomeSize++;
	}
	
	/**
	 * decrements the polyosome size
	 *
	 */
	public void decrementPolyosomeSize(double time){
		this.polyosomeSize--;
	}
	
	/**
	 * checks whether the system prints the polyosome size
	 * @return
	 */
	public boolean printsPolyosomeSize(){
		return printPolyosomeSize;
	}
	
	public void printPolyosomeSize(String filename) throws FileNotFoundException{
		//print mRNA types
		String result="";
		result+="#time, size, relative/mRNA, relative/codon, relative/maximum\n";
		String str="";
		double bufferRelmRNA, bufferRelCodon, bufferRelMax;
		//Utils.writeToFile(filename,"0, 0, 0",true);
		int totalLength=0;
		for (int i=0;i<mRNA.length;i++){
			totalLength+=mRNA[i].getLength();
		}
		
		int totalNoOfPositions  = (int) (3 + Math.floor(totalLength-(3+this.noOfRibosomeSitesAfter + this.noOfRibosomeSitesBefore))/12);
		
		for(int i=0;i<ps.size();i++){
			bufferRelmRNA = (double) ps.get(i).getData()/mRNA.length;
			bufferRelCodon = (double) ps.get(i).getData()/totalLength;
			bufferRelMax = (double) ps.get(i).getData()/totalNoOfPositions;
			
			str = ps.get(i).toString() + ", "+bufferRelmRNA + ", "+bufferRelCodon + ", "+bufferRelMax+"\n";
			result+=str;
		}
		
		Utils.writeToFile(filename,result,true);
	}
	
	/**
	 * checks whether the system is in debug mode
	 * @return
	 */
	public boolean isInDebugMode(){
		return debugMode;
	}
	
	/**
	 * checks whether the system is exact or approximate
	 * @return
	 */
	public boolean isApproximateSystem(){
		return approximateSystem;
	}
	
	/**
	 * returns true if elongation binding uses a Gillespie Direct algorithm
	 * @return
	 */
	public boolean doesElongationBindingEventsUseGillespeDirect(){
		return this.elongationBindingEventsUseGillespeDirect;
	}
	
	/**
	 * returns true if non-cognate introduce waiting time
	 * @return
	 */
	public boolean doesNonGognateIntroduceWaitingTime(){
		return this.nonGognateIntroduceWaitingTime;
	}
	
	/**
	 * returns true if the non-cognate affect the binding rate
	 * @return
	 */
	public boolean  doesNonCognateChangeBindingRate(){
		return this.nonCognateChangeBindingRate;
	}
	
	
	
	/**
	 * returns the waiting time per non-cognate binding attempt
	 * @return
	 */
	public double getNearGognateWaitingTime(){
		return this.nearGognateWaitingTime;
	}
	/**
	 * returns the waiting time per non-cognate binding attempt
	 * @return
	 */
	public double getNonGognateWaitingTime(){
		return this.nonGognateWaitingTime;
	}
	
	/**
	 * returns the Michaelis-Menten constant of how the binding rate of (near-)cognates is changed depending on non-cognate abundance 
	 * @return
	 */
	public double getNonCognateMMConstant(){
		return this.nonCognateMMConstant;
	}
	
	/**
	 * checks whether the ribosome can unbind from IBS if free
	 * @return
	 */
	public boolean unbindsFromIBSifFree(){
		return unbindsFromIBSifFree;
	}
	
	/**
	 * checks whether the ribosome can unbind from termination if free
	 * @return
	 */
	public boolean unbindsFromterminationIfFree(){
		return unbindsFromterminationIfFree;
	}
	
	/**
	 * cheks whether a ribosome unbinds if stuck
	 * @return
	 */
	public boolean  unbindsFromCodonIfBusy(){
		return unbindsFromCodonIfBusy;
	}
	
	
	
	/**
	 * checks whether a ribosome can move or not
	 * if it cannot then it sets the blocked by field with 
	 * @param ribosomeID
	 * @return the next position to move to
	 */
	public int canMove(int ribosomeID){
		int buffer =  ribosomes[ribosomeID].getBlockedBy();
		
		if(buffer!=Constants.NONE){
			if(ribosomes[ribosomeID].getmRNAPosition() == Constants.IBS_POSITION){
				ribosomes[buffer].setBlocksIBS(Constants.NONE);
			} else{
				ribosomes[buffer].setBlocks(Constants.NONE);
			}
		}
		
		int result = ribosomes[ribosomeID].canTranslocate(mRNA[ribosomes[ribosomeID].getmRNA()], this.noOfRibosomeSitesAfter, this.noOfRibosomeSitesBefore, this.stallEventsWait);
		
		if(ribosomes[ribosomeID].getBlockedBy() != Constants.NONE){
			if(ribosomes[ribosomeID].getmRNAPosition() == Constants.IBS_POSITION){
				ribosomes[ribosomes[ribosomeID].getBlockedBy()].setBlocksIBS(ribosomeID);
			} else{
				ribosomes[ribosomes[ribosomeID].getBlockedBy()].setBlocks(ribosomeID);
			}
		}
		
		return result;
	}
	
	/**
	 * returns the charged abundance of a specific tRNA species.
	 * @param tRNAID
	 * @return
	 */
	public final  double getTRNAChargedAbundance(int tRNAID){
		return this.tRNATabe.gettRNAChargeAbundance(tRNAID);
	}
	
	/**
	 * checks whether the mRNA copy number is stochastic
	 * @return
	 */
	public boolean stochasticmRNACopyNumber(){
		return stochasticmRNACopyNumber;
	}
	
	/**
	 * returns the elongation constant rate for cognates
	 * @return
	 */
	public double getElongationCognateRate(){
		return this.elongationCognateRate;
	}
	
	
	/**
	 * returns the elongation constant probability  for cognates
	 * @return
	 */
	public double getElongationCognateProbability(){
		return this.elongationCognateProbability;
	}
	
	/**
	 * returns the elongation constant rate  for near-cognates
	 * @return
	 */
	public double getElongationNearCognateRate(){
		return this.elongationNearCognateRate;
	}
	
	/**
	 * returns the elongation constant probability  for near-cognates
	 * @return
	 */
	public double getElongationNearCognateProbability(){
		return this.elongationNearCognateProbability;
	}
	
	
	/**
	 * sets which ribosome is blocked by the current one
	 * @param ribosomeID
	 */
	public void setBlocks(int ribosomeID, int blocksID){
		ribosomes[ribosomeID].setBlocks(blocksID); 
	}
	
	
	/**
	 * sets which ribosome is blocked by the current one at IBS
	 * @param ribosomeID
	 */
	public void setBlocksIBS(int ribosomeID, int blocksID){
		ribosomes[ribosomeID].setBlocksIBS(blocksID); 
	}
	
	/**
	 * sets which ribosome blocks the current one
	 * @param ribosomeID
	 */
	public void setBlockedBy(int ribosomeID, int blockedByID){
		ribosomes[ribosomeID].setBlockedBy(blockedByID); 
	}
	
	
	/**
	 * gets which ribosome is blocked by the current one
	 * @param ribosomeID
	 */
	public int getBlocks(int ribosomeID){
		return ribosomes[ribosomeID].getBlocks(); 
	}
	
	/**
	 * gets which ribosome is blocked by the current one
	 * @param ribosomeID
	 */
	public int getBlocksIBS(int ribosomeID){
		return ribosomes[ribosomeID].getBlocksIBS(); 
	}
	
	/**
	 * gets which ribosome blocks the current one
	 * @param ribosomeID
	 */
	public int getBlockedBy(int ribosomeID){
		return ribosomes[ribosomeID].getBlockedBy(); 
	}
	
	/**
	 * returns true if the ribosome is at termination site
	 * @param ribosomeID
	 * @return
	 */
	public boolean isAtTerminationSite(int ribosomeID){
		return mRNA[ribosomes[ribosomeID].getmRNA()].getLength() - 1 == ribosomes[ribosomeID].getmRNAPosition();
		
	}
	
	/**
	 * returns true if the ribosome is at the IBS site
	 * @param ribosomeID
	 * @return
	 */
	public boolean isAtIBSSite(int ribosomeID){
		return Constants.IBS_POSITION == ribosomes[ribosomeID].getmRNAPosition();
		
	}
	
	/**
	 * returns the translocation rate
	 * @return
	 */
	public double getElongationTranslocationRate(){
		return this.elongationTranslocationRate;
	}
	
	/**
	 * mark an mRNA strand as occupied at at specified position
	 * @param mRNAID the id of the mRNA
	 * @param pos the position
	 */
	public void occupymRNAPosition(int mRNAID, int pos, int ribosomeID){
		mRNA[mRNAID].setOccupancyState(pos, ribosomeID);
	}
	
	/**
	 * updates the cell time
	 * @param time
	 */
	public void updateCellTime(double time){
		cellTime = time;
	}
	
	/**
	 * returns an mRNA
	 * @param id
	 * @return
	 */
	public MessengerRNA getmRNA(int id){
		return mRNA[id];
	}
	
	/**
	 * increment the number of successful elongations
	 *
	 */
	public void incrementSuccessfulElongation(double time){
		if(time > this.recordStartTime){
			this.succesfullElongations++;
		}
	}
	
	
	/**
	 * returns the length of the specified mRNA
	 * @param mRNAID
	 * @return
	 */
	public int getmRNAlength(int mRNAID){
		return mRNA[mRNAID].getLength();
	}
	
	/**
	 * returns the codon at a specified position on a specific mRNA
	 * @param mRNAID the specific mRNAID
	 * @param pos the position of the codon
	 * @return the codon
	 */
	public byte getCodon(int mRNAID, int pos){
		return mRNAtype[mRNA[mRNAID].getType()].getCodon(pos);
	}
	
	
	
	/**
	 * increment the number of unsuccessful elongations
	 *
	 */
	public void incrementUnSuccessfulElongation(double time){
		if(time > this.recordStartTime){
			this.unsuccesfullElongations++;
		}
	}
	
	/**
	 *  slides a ribosome on the mRNA by changing the mRNA occupancy state
	 * @param mRNAID
	 * @param originalPosition
	 * @param nextPosition
	 */
	public void slideRibosomeOnmRNA(int mRNAID, int originalPosition, int nextPosition){
		mRNA[mRNAID].slide(originalPosition, nextPosition);
	}
	
	
	/**
	 * increments the number of translated mRNAs for the specified species
	 * @param mRNAID
	 */
	public void incrementTranslated(int mRNAID, double time){
		if(time > this.recordStartTime){
			mRNATranslated[mRNA[mRNAID].getType()]++;
		}
	}

//	/**
//	 * increments the number of free bound ribosomes
//	 */
//	public void incrementNumberOfFreeBoundRibosomes(){
//		this.numberOfFreeBoundRibosomes++;
//	}
//	
//
//	/**
//	 * decrements the number of free bound ribosomes
//	 */
//	public void decrementNumberOfFreeBoundRibosomes(){
//		this.numberOfFreeBoundRibosomes--;
//	}
//	
//	/**
//	 * returns the number of ribosome waiting for the tRNA to bind
//	 * @return
//	 */
//	public int getNumberOfFreeBoundRibosomes(){
//		return this.numberOfFreeBoundRibosomes;
//	}

	public int getmRNAType(int mRNAID){
		return mRNA[mRNAID].getType();
	}
	
	
	/**
	 * prints a mRNA type to a string
	 * @param id the id of the mRNA
	 * @param codons whether to print literal codons or just numbers
	 * @return the string of the mRNA
	 */
	private String getmRNAType(int id, boolean codons){
		String result="";
		
		if(id>=0 && id<mRNAtype.length){
			result= mRNAtype[id].getName()+", ";
			int length =mRNAtype[id].getLength();
			result+= length+", " + mRNAtype[id].getRibosomeAffinity()+", "+ mRNAtype[id].getCopyNumber()+", ";
			byte buffer;
			if(codons){
				for(int i=0;i<length;i++){
					buffer = mRNAtype[id].getCodon(i);
					result+=SequenceUtils.codons.codon[buffer]+" ";
				}
			} else{
				for(int i=0;i<length;i++){
					buffer = mRNAtype[id].getCodon(i);
					result+=buffer+" ";
				}
			}
			
		}
		
		return result;
	}
	
	
	private String mRNAtoString(int id){
		String result="";
		
		if(id>=0 && id<mRNA.length){
			result= "id: "+id+", "+mRNA[id].getName()+",   ";
			int length =mRNA[id].getLength();
			result+= length+",   ";
			byte buffer;
			for(int i=0;i<length;i++){
				buffer = 0;
				if(mRNA[id].isOccupied(i)!=Constants.NONE){
					buffer=1; 
				}
				result+=buffer+" ";
			}
			
		}
		
		return result;
	}
	

	public void printmRNAs() throws FileNotFoundException{
		printStateToFile(this.resultsFilepath+this.resultsFilename+".txt", false,false);
	}
	
	/**
	 * prints to a file the cell population
	 * @param filename the filename where to write
	 * @param codons true if codons are writtent explicitelly or false if a number is used instead
	 * @param append true if to append the file
	 * @throws FileNotFoundException
	 */
	public void printStateToFile(String filename, boolean codons,boolean append) throws FileNotFoundException{
	
		//print mRNA types
		printDebugInfo("print mRNA types");
		Utils.writeToFile(filename,"=======================================\n"+"#mRNA types",append);
		for(int i=0;i<mRNAtype.length;i++){
			Utils.writeToFile(filename,getmRNAType(i,codons),true);
		}
		Utils.writeToFile(filename,"=======================================\n",true);
		
	}	
	

	
		/**
	 * removes a ribosome from the free list and puts it in the occupied list
	 * @param ribosomeID the id of the ribosome geting occupied
	 */
	public boolean occupyRibosome(int ribosomeID){
	int index =freeRibosomes.size()-1;
        while(index>=0){
        
            if(freeRibosomes.get(index)==ribosomeID){
	
            freeRibosomes.remove(index); 
			occupiedRibosomes.add(ribosomeID);
                       
			return true;
                                            }
        index--;
		} 
	
        System.out.println("I suspect we should never be in Cell:occupyRibosome llkas" + ribosomeID); 
        System.exit(-1);
        return false;
	}
	
	
	
 

	/**
	 * removes a mRNA from the free list and puts it in the occupied list 
	 * @param mRNAID the id of the mRNA
	 */

       public boolean occupymRNA(int mRNAID){
               boolean done=false;
               int mRNATypeID = this.mRNA[mRNAID].getType();
               for (int i=0; i< freemRNAs.get(mRNATypeID).size();i++){
                       if(freemRNAs.get(mRNATypeID).get(i)==mRNAID){
                               freemRNAs.get(mRNATypeID).remove(i);
                               //totalFreeMRNAs--;
                               done=true;
                       }
               }
               return done;
       }




	
	 

	/**
	 * sets a charged tRNA to an busy one (bound to the ribosome)
	 * @param  tRNAID the id of the tRNA species that is affected
	 * @return true if the charged abundance is higher than 0 or false otherwise
	 */
	public boolean settRNABusy(int tRNAID){
		return tRNATabe.setBusy(tRNAID);
	}
	
	
	
	
	/**
	 * sets a busy (bound to the ribosome) tRNA to an charged one 
	 * @param  tRNAID the id of the tRNA species that is affected
	 * @return true if the busy abundance is higher than 0 or false otherwise
	 */
	public boolean resettRNABusy(int tRNAID){
		return tRNATabe.resetBusy(tRNAID);
	}
	
	
	/**
	 * sets a busy (bound to the ribosome) tRNA to an uncharged one 
	 * @param  tRNAID the id of the tRNA species that is affected
	 * @return true if the busy abundance is higher than 0 or false otherwise
	 */
	public boolean unchargetRNA(int tRNAID){
		boolean uncharged = tRNATabe.uncharge(tRNAID);
		return uncharged;
	}
	
	/**
	 *  Schedules the next tRNA charging event
	 * @param tRNAID the tRNA species
	 * @param cellTime the current time
	 */
	public void scheduleNexttRNAEvent(int tRNAID,double cellTime){
		eventQueue.scheduletRNACharging(tRNAID, cellTime, this.tRNATabe.gettRNAchargeRate(tRNAID), this.randomGenerator);
	}
	
	/**
	 * returns the tRNA binding rate
	 * @return
	 */
	public double gettRNAbindingRate(){
		return this.tRNAbindingRate;
	}
	
        
        
        
      
	/**
	 * removes a ribosome from the free list and puts it in the occupied list
	 * @param ribosomeID the id of the ribosome geting occupied
	 * @return if removed the ribosome from occupied
	 */
	public boolean freeRibosome(int ribosomeID){
		if( occupiedRibosomes.remove(ribosomeID)){
			freeRibosomes.add(ribosomeID);
			return true;
		}
		return false;
	}
	
	/**
	 * removes a mRNA from the free list and puts it in the occupied list 
	 * @param mRNAID the id of the mRNA
	 */
	public void freemRNA(int mRNAID){
		freemRNAs.get(mRNA[mRNAID].getType()).add(mRNAID);
	}
	
	/**
	 * performs the unbinding of a ribosome from the mRNA
	 * @param ribosomeID the ID of the ribosome that unbinds
	 * @return  a 2 dim array of int 1) -1 if ribosomes remains 
         * bound or the id of the ribosome if this gets unbound 
         * 2) the ID of the mRNA if the mRNA IBS site gets free or -1 otherwise
	 */
	public FreedAgents releaseRibosome(int ribosomeID,double time){
		FreedAgents gotFree=new  FreedAgents();
		int originalPosition = ribosomes[ribosomeID].getmRNAPosition();
		int mRNAID = ribosomes[ribosomeID].getmRNA();
		// perform unbinding
		 
		
		boolean unbided = ribosomes[ribosomeID].unbind(mRNA[mRNAID]);
		if(unbided){
			
			//if(tRNAIDsiteP!=Constants.NONE){
			//	
			//}
			
			
			mRNA[mRNAID].setOccupancyState(originalPosition, Constants.NONE);
			//update occupancy lists
			freeRibosome(ribosomeID);
			if(originalPosition == 0){ 
				freemRNA(mRNAID);
				gotFree.mRNA = mRNAID;
			}
			//
			gotFree.ribosome = ribosomeID;
			this.decrementRibosomesPerMRNA(mRNAID, time);
			
			//remove the interconnections
			if(ribosomes[ribosomeID].getBlockedBy()!=Constants.NONE){
				if(ribosomes[ribosomeID].getmRNAPosition()==Constants.IBS_POSITION){
					setBlocksIBS(ribosomes[ribosomeID].getBlockedBy(), Constants.NONE);
				} else{
					setBlocks(ribosomes[ribosomeID].getBlockedBy(), Constants.NONE);
				}
				setBlockedBy(ribosomeID,Constants.NONE);
			}
			
			if(ribosomes[ribosomeID].getBlocks()!=Constants.NONE){
				setBlockedBy(ribosomes[ribosomeID].getBlocks(), Constants.NONE);
				setBlocks(ribosomeID,Constants.NONE);
			}
			
			if(ribosomes[ribosomeID].getBlocksIBS()!=Constants.NONE){
				setBlockedBy(ribosomes[ribosomeID].getBlocksIBS(), Constants.NONE);
				setBlocksIBS(ribosomeID,Constants.NONE);
			}
                        
            //This is only relevant when there were no previous ribosomes free.
            //now after un-binding there is, and hence we need to re-schedule a new
            //event of ribo binding to mRNA
                  if(!Ribosome.existFreeRibos()) {
                        Ribosome.setIsFreeRibos(true);      
                      //if there was no free  ribo left, schedule a new binding event
                    
                    this.eventQueue.scheduleRibosomeBinding(this);                                       
                      
                  
                                                                                }
		  
                }
		

		return gotFree;
	}


        
        
        
        
	/**
	 * returns the last ribosome
	 */
	public int getLastFreeRibosome(){
            
            int size=freeRibosomes.size();
            
            if(size>0)
		return this.freeRibosomes.get(freeRibosomes.size()-1);
            else return -1;
            
        }
	
	/**
	 * returns the last free mRNA of a specific type 
	 * @param mRNAID the id of the mRNA type
	 * @return
	 */
	public int getLastFreemRNA(int mRNAID){
		return freemRNAs.get(mRNAID).get(freemRNAs.get(mRNAID).size()-1);
	}
	
	/**  
	*  ADDED BY CHU to replace getLastFreeRNA
	 * returns the random free mRNA of a specific type 
	 * @param mRNAID the id of the mRNA type
	 * @return
	 */
	public int getRandomFreemRNA(int mRNAID){
	int maxNum =freemRNAs.get(mRNAID).size();
	int test;
      
        int lustig =  randomGenerator.nextInt(maxNum) 	;
	//System.out.println( " LUSTIG ---> " + lustig + "   " +  mRNA[freemRNAs.get(mRNAID).get(lustig)] )	;
	return freemRNAs.get(mRNAID).get(lustig);
	}


	/**
	 * returns the mRNA where the ribosome is bound
	 * @param ribosomeID
	 * @return
	 */
	public int getRibosomemRNA(int ribosomeID){
		return ribosomes[ribosomeID].getmRNA();
	}
	
	/**
	 * returns the position of a ribosome on the mRNA
	 * @param ribosomeID
	 * @return
	 */
	public int getRibosomePositionOnmRNA(int ribosomeID){
		return ribosomes[ribosomeID].getmRNAPosition();
	}
	
	/**
	 * for e mRNA get the random walk rate on the IBS
	 * @param mRNAID
	 * @return
	 */
	public double getmIBSMoveRate(int mRNAID){
		return mRNAtype[mRNA[mRNAID].getType()].getIBSmoveRate();
	}
	
	/**
	 * returns current state of a ribosome
	 * @param ribosomeID
	 * @return
	 */
	public int getRibosomeState(int ribosomeID){
		return ribosomes[ribosomeID].getState();
	}
	
	/**
	 * gets the current codon where a ribosome is located
	 * @param ribosomeID
	 * @return
	 */
	public byte getRibosomeCodon(int ribosomeID){
		return ribosomes[ribosomeID].getCodon();
	}
	
	/**
	 * gets the current tRNA bound to the ribosome
	 * @param ribosomeID
	 * @return
	 */
	public int getRibosometRNA(int ribosomeID){
		return ribosomes[ribosomeID].gettRNAIDsiteA();
	}
	
	/**
	 * checks whether the bound tRNA is cognate or not
	 * @param ribosomeID
	 * @return
	 */
	public boolean isRibosometRNACognate(int ribosomeID){
		return ribosomes[ribosomeID].isCognate();
	}
	
	/**
	 * checks whether a ribosome is bound to the mRNA or not?
	 * @param ribosomeID
	 * @return
	 */
	public boolean isRibosomeFree(int ribosomeID){
		return ribosomes[ribosomeID].getmRNA()< 0;
	}
	
	/**
	 * returns the next position of the tRNA
	 * @param ribosomeID
	 * @return
	 */
	public int getRibosomeNextmRNAPosition(int ribosomeID){
		return ribosomes[ribosomeID].getNextPosition(mRNA[ribosomes[ribosomeID].getmRNA()].getLength());
	}
	/**
	 * returns the last reaction time of a ribosome
	 * @param ribosomeID
	 * @return
	 */
	public double getRibosomeLastReactionTime(int ribosomeID){
		return ribosomes[ribosomeID].getLastReactionTime();
	}
	
	/**
	 * gets the total number of free ribosomes
	 * @return
	 */
	public int getNoOfFreeRibosomes(){
		return freeRibosomes.size();
	}
	
        public double getFractionOfFreeRibosomes(){
            
	double b= (double)freeRibosomes.size();
        
        double c= (double) noOfRibosomes;
        return b/c;
	}
	
        
        
	/**
	 * returns the tRNA charged abundances
	 * @param codonID
	 * @return
	 */
	public TRNAPropensity gettRNAChargedAbundances(byte codonID){
		return this.tRNATabe.getChargedAbundances(codonID);
	}
	
	/**
	 * returns the total tRNA charged abundance
	 * @return
	 */
	public int gettRNATotalChargedAbundances(){
		return this.tRNATabe.getChargedAbundance();
	}
	/**
	 * schedules a ribosome event
	 * @param ribosomeID the ID of the ribosome to be schduled
	 * @param cellTime the cell time
	 * @param gotFree 
	 */
/*	public void scheduleNextRibosomeEvent(int ribosomeID, double cellTime){
//		eventQueue.scheduleNextRibosomeEvent(ribosomeID, cellTime,  this, randomGenerator);
	}*/
	
  
	/**
	 * updates the next events that are affected by a change in the tRNA population
	 * @param tRNAID the tRNA that changes abundance
	 */
	//public void updateElongationEventQueueBindingCodon(int tRNAID, double cellTime){
	//	eventQueue.updateElongationEventBinding(tRNAID, cellTime, this, this.randomGenerator);
//	}

	public void run(double stopTime)  throws FileNotFoundException{
		this.stopTime = stopTime;
		run();
	}
	
	
	
	
	

	/**
	 * run the simulations for a number of time steps
	 * @param stopTime the end time of the simulations.
	 */
	public void run()  throws FileNotFoundException{

	
		 
		Event e;
		RibosomeEvent re;
		TRNAEvent te;
		FreedAgents gotFree=new FreedAgents(); 
		
		long curTime = System.currentTimeMillis();
		int nextProcent=10;
		double elapsedTimeSec;
		int ribosomeID;
		boolean toScheduleRibosomeBinding;
		
		//System.out.println(stallEventsWait);
		//set simulation time to 0
		cellTime=0;
		
		// schedule first binding event
		toScheduleRibosomeBinding = false;
		//schedule the first binding event
	//eventQueue.scheduleNextRibosomeEvent(this.getLastFreeRibosome(),this.cellTime,this, randomGenerator);
       eventQueue.scheduleRibosomeBinding(this);
        
		if(eventQueue.size() >0){
			toScheduleRibosomeBinding = true;
		}

		//int count = 0;
		
		double stepPolyosomeSizePoint = (double) stopTime/this.polyosomeSizePrintPoints;
		double nextPolyosomeSizePoint = stepPolyosomeSizePoint;
		
		
		while (cellTime < this.stopTime){
			gotFree=new FreedAgents(); 

//       int a=getNumberOfFreeBoundRibosomes();
//       int b= getNoOfFreeRibosomes();
//       int c= a+b;
//       System.out.println(a+b + "  "+ a + "  "+ b);
       
			e=null;
			e=eventQueue.getNextEvent();
			if(e != null){
				// there is at least one event in the queue
                                    cellTime=e.time;
				if(cellTime > this.stopTime){System.out.println("Halted at time: " + cellTime);}	
				if(e instanceof RibosomeEvent){
				
                                    // next event is a ribosome event
					re = (RibosomeEvent)e;
					ribosomeID = re.ribosomeID;
                                        
          
                                        gotFree = this.ribosomes[ribosomeID].act(re, this);
			 
				} else if( e instanceof TRNAEvent){
					// next event is a tRNA event 
					te = (TRNAEvent) e;
					this.tRNATabe.act(te, this);
				}
			} else{
				System.out.println("hm... no more events!");
				break;
			}
			
			if(!debugMode && nextProcent <=  100*cellTime/stopTime){
				elapsedTimeSec = (double) (System.currentTimeMillis()-curTime)/1000;
				printDebugInfo(nextProcent+"% finished in "+ elapsedTimeSec + " seconds");
			
				nextProcent = nextProcent+10;
			}
			
			if(this.printPolyosomeSize && nextPolyosomeSizePoint<=cellTime){
				ps.add(new DynamicBehaviour<Integer>(cellTime,this.polyosomeSize));
				nextPolyosomeSizePoint+=stepPolyosomeSizePoint;
			}

		}
 
//Print out diagnostic information
System.out.println(" Time: " + cellTime + " free mRNA: " + getFreemRNACopyNumber() + " NO OF FREE RIBOS  " +  getNoOfFreeRibosomes()  + "  total number charged tRNA:  "+ gettRNATotalChargedAbundances() + "  "     );			
	}
	
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
	/**
	 * load parameters from a specific file 
	 * @param filename the file from where the params are loaded
	 */
	private void loadParametersFromFile(String filename){
		
		if(!filename.isEmpty()){
			//System.out.println(filename);
				BufferedReader reader = null;
			    try{
		            reader = new BufferedReader(new FileReader(filename));
		            String text = null;
		            while ((text = reader.readLine()) != null)
		            {
		            	text=text.trim();
		            	if(!text.isEmpty() && !text.startsWith("#")){
		            		if(text.endsWith(";")){
		            			setFileParameter(text);
		            		} else{
		            			printDebugInfo("Error interpretting line:"+text,false);
		            		}
		            	} 
		            }
		            //printDebugInfo("Cell created with parameters from file "+filename,false);
		        } catch (Exception e) {
	    			//printDebugInfo("Cell created with default parameters",false);	
	    			printDebugInfo(e.toString(),true);
	            } finally {
	                try{
	                    if (reader != null){
	                        reader.close();
	                    }
	                } catch (IOException e){
	                	printDebugInfo(e.toString(),true);
	                }
	           	}  
			}
    			
	}
	
	
	
	/**
	 * takes one command line and interprets it
	 * @param text the command line
	 * @return true if the parameter was found and changed.
	 */
	private boolean setFileParameter(String text){
		boolean result=false;
		String[] param = extractParameterFromCommandLine(text);
		if(param.length==2 && !param[0].isEmpty()){
			result=true;
			//System.out.println(param[0] + " = " + param[1]);
			if(param[0].equals("STOP_TIME")){
				this.stopTime = Double.parseDouble(param[1]);
			} else if(param[0].equals("RECORD_START_TIME")){
				this.recordStartTime = Double.parseDouble(param[1]);
			} else if(param[0].equals("NO_OF_MRNA")){
				this.noOfmRNAtype = Integer.parseInt(param[1]);
			} else if(param[0].equals("NO_OF_COPIES_PER_MRNA_AVG")){
				this.noOfCopiesPermRNAAvg = Integer.parseInt(param[1]);
			} else if(param[0].equals("MRNA_CODON_LENGTH_MIN")){
				this.mRNAcodonLengthMin = Integer.parseInt(param[1]);
			} else if(param[0].equals("MRNA_CODON_LENGTH_MAX")){
				this.mRNAcodonLengthMax = Integer.parseInt(param[1]);
			} else if(param[0].equals("RANDOM_MRNA_SEED")){
				this.randommRNASeed = Long.parseLong(param[1]);
			} else if(param[0].equals("MRNA_FILENAME")){
				//System.out.println(param[0] + " = " + param[1]);
				this.mRNAfilename = param[1];	
			} else if(param[0].equals("MRNA_ABUNDANCE_FILENAME")){
				this.mRNAabundanceFilename = param[1];	
			} else if(param[0].equals("IBS_LENGTH_AVG")){
				this.IBSlengthAvg =Double.parseDouble(param[1]);	
			} else if(param[0].equals("IBS_SCANNING_RATE")){
				this.IBSscanningRate =Double.parseDouble(param[1]);	
			} else if(param[0].equals("NO_OF_COPIES_PER_MRNA_MIN")){
				this.noOfCopiesPermRNAMin =  Double.parseDouble(param[1]);
			} else if(param[0].equals("NO_OF_SYNTHESISED_PROTEINS_PER_MRNA_MIN")){
				this.noOfSynthesisedProteinsPermRNAmin = Double.parseDouble(param[1]);
			} else if(param[0].equals("NO_OF_RIBOSOMES")){
				this.noOfRibosomes = Integer.parseInt(param[1]);
			} else if(param[0].equals("RATIO_OF_FREE_RIBOSOMES")){
				this.ratioOfFreeRibosomes = Double.parseDouble(param[1]);
			} else if(param[0].equals("FILE_PATH")){
				this.resultsFilepath = param[1];
			} else if(param[0].equals("FILENAME")){
				this.resultsFilename = param[1];
			} else if(param[0].equals("OUTPUT_FILE")){
				this.debugOutputFile = param[1];
			} else if(param[0].equals("DEBUG_MODE")){ 
				this.debugMode= Boolean.parseBoolean(param[1]);
			} else if(param[0].equals("APPROXIMATE_SYSTEM")){
				this.approximateSystem= Boolean.parseBoolean(param[1]);
			} else if(param[0].equals("UNBINDS_FROM_IBS_IF_BUSY")){
				this.unbindsFromIBSifFree= Boolean.parseBoolean(param[1]);
			} else if(param[0].equals("UNBINDS_FROM_TERMINATION_IF_BUSY")){
				this.unbindsFromterminationIfFree= Boolean.parseBoolean(param[1]);
			} else if(param[0].equals("UNBINDS_FROM_CODON_IF_BUSY")){
				this.unbindsFromCodonIfBusy = Boolean.parseBoolean(param[1]);
			} else if(param[0].equals("COLLISIONS_BEFORE_UNBINDS")){
				this.collisionsBeforeUnbinds =Integer.parseInt(param[1]);
			} else if(param[0].equals("STOCHASTIC_MRNA_COPY_NUMBER")){  
				this.stochasticmRNACopyNumber = Boolean.parseBoolean(param[1]);
			} else if(param[0].equals("POLYOSOME_SIZE_PRINT_POINTS")){
				this.polyosomeSizePrintPoints = Integer.parseInt(param[1]);
				this.printPolyosomeSize = false;
				if(polyosomeSizePrintPoints > 0){
					this.printPolyosomeSize = true;
				}			
			} else if(param[0].equals("NON_COGNATE_INTRODUCE_WAITING_TIME")){
				this.nonGognateIntroduceWaitingTime = Boolean.parseBoolean(param[1]);

			} else if(param[0].equals("NEAR_COGNATE_WAIT_TIME")){
				//System.out.println(param[0] + " = " + param[1]);
				this.nearGognateWaitingTime = Double.parseDouble(param[1]);
	

			} else if(param[0].equals("NON_COGNATE_WAIT_TIME")){
				//System.out.println(param[0] + " = " + param[1]);
				this.nonGognateWaitingTime = Double.parseDouble(param[1]);
			


} else if(param[0].equals("NON_COGNATE_CHANGE_BINDING_RATE")){
				//System.out.println(param[0] + " = " + param[1]);
				this.nonCognateChangeBindingRate = Boolean.parseBoolean(param[1]);
			} else if(param[0].equals("TRNA_ABUNDANCE_FILENAME")){
				this.tRNAAbundaceFile = param[1];
			} else if(param[0].equals("TRNA_BINDING_RATE")){
				this.tRNAbindingRate = Double.parseDouble(param[1]);
			} else if(param[0].equals("TRNA_CHARGE_RATE")){
				this.tRNAchargeRate = Double.parseDouble(param[1]);
			} else if(param[0].equals("RIBOSOME_AFFINITY_AVG")){
				this.ribosomeAffinityStd = Double.parseDouble(param[1]); 
				this.ribosomeAffinityAvg = Double.parseDouble(param[1]);
			} else if(param[0].equals("RIBOSOME_AFFINITY_STD")){
				//this.ribosomeAffinityStd = Double.parseDouble(param[1]); 
			} else if(param[0].equals("IBS_MOVE_RATE_MIN")){
				this.IBSmoveRateMin = Double.parseDouble(param[1]);
			} else if(param[0].equals("IBS_MOVE_RATE_MAX")){
				this.IBSmoveRateMax = Double.parseDouble(param[1]);
			} else if(param[0].equals("NO_OF_RIBOSOME_SITES_BEFORE")){
				this.noOfRibosomeSitesBefore = Integer.parseInt(param[1]);
			} else if(param[0].equals("NO_OF_RIBOSOME_SITES_AFTER")){
				this.noOfRibosomeSitesAfter = Integer.parseInt(param[1]);
			} else if(param[0].equals("NO_OF_RIBOSOME_SITES_TOTAL")){
				this.noOfRibosomeSitesTotal = Integer.parseInt(param[1]);
			} else if(param[0].equals("NO_OF_TRANSLATION_PER_RIBOSOME_MRNA_MAX")){
				this.noOfTranslationPerRibosomemRNAMax = Integer.parseInt(param[1]);
			} else if(param[0].equals("NO_OF_TRANSLATION_PER_RIBOSOME_MRNA_MIN")){
				this.noOfTranslationPerRibosomemRNAMin = Integer.parseInt(param[1]);
			} else if(param[0].equals("ELONGATION_COGNATE_PROBABILITY")){
				this.elongationCognateProbability = Double.parseDouble(param[1]);
			} else if(param[0].equals("ELONGATION_NEARCOGNATE_PROBABILITY")){
				this.elongationNearCognateProbability = Double.parseDouble(param[1]);
			} else if(param[0].equals("ELONGATION_NEARCOGNATE_PROBABILITY_EXIT_RECOGNITION")){
				this.elongationNearCognateProbabilityExitRecognition = Double.parseDouble(param[1]);				
			} else if(param[0].equals("ELONGATION_COGNATE_PROBABILITY_EXIT_RECOGNITION")){
				this.elongationCognateProbabilityExitRecognition = Double.parseDouble(param[1]);				
			} else if(param[0].equals("ELONGATION_COGNATE_RATE")){
				this.elongationCognateRate = Double.parseDouble(param[1]);
			} else if(param[0].equals("ELONGATION_NEARCOGNATE_RATE")){
				this.elongationNearCognateRate = Double.parseDouble(param[1]);
			} else if(param[0].equals("ELONGATION_TRANSLOCATION_RATE")){
				this.elongationTranslocationRate = Double.parseDouble(param[1]);
			} else if(param[0].equals("ELONGATION_NONCOGNATE_PROBABILITY")){
				this.elongationNonCognateProbability = Double.parseDouble(param[1]);
			} else if(param[0].equals("ELONGATION_BINDING_EVENTS_USE_GILESPIE_DIRECT")){
				this.elongationBindingEventsUseGillespeDirect = Boolean.parseBoolean(param[1]);
			} else if(param[0].equals("STALL_EVENTS_WAIT")){
				this.stallEventsWait = Boolean.parseBoolean(param[1]);
			} else{
				printDebugInfo("Unknown parameter: "+text);
				 result=false;
			}
			
			
		}
		
		return result;
	}
	
	/**
	 * breaks the command line into two strings the parameter and the value
	 * @param text the command line
	 * @return a 2 value array of strings 0-the parameter 1- the value
	 */
	private String[] extractParameterFromCommandLine(String text){
		String[] result= new String[2];
		int firstPos, lastPos;
		result[0]="";
		result[1]="";
		if(text.contains("=")){
			firstPos = text.indexOf("=");
			lastPos = text.lastIndexOf("=");
			if(firstPos >=0 && firstPos == lastPos){
				result[0]=text.substring(0,firstPos);
				//remove spaces
				result[0] = result[0].trim();
				result[1]=text.substring(firstPos+1,text.length()-1);
				//remove spaces
				result[1] = result[1].trim();
				//remove "
				result[1] = result[1].replace("\"", "");
			} else{
				printDebugInfo("error in parameters line: \""+text+"\" (more than one equal signs)");
			}
		} else{
			printDebugInfo("error in parameters line: \""+text+"\" (no equal sign)");
		}

		
		return result;
	}
	
	/**
	 * sets the default values for the parameters provided from outside 
	 */
	private void initialiseInputParameters(){
		this.stopTime = Constants.STOP_TIME;
		this.noOfmRNAtype = Constants.NO_OF_MRNA;
		this.noOfCopiesPermRNAAvg = Constants.NO_OF_COPIES_PER_MRNA_AVG;
		this.mRNAcodonLengthMin = Constants.NO_OF_CODONS_PER_MRNA_MIN;
		this.mRNAcodonLengthMax = Constants.NO_OF_CODONS_PER_MRNA_MAX;
		this.randommRNASeed = Constants.RANDOM_MRNA_SEED;
		this.mRNAfilename = Constants.MRNA_FILENAME;
		this.noOfRibosomes = Constants.NO_OF_RIBOSOMES;
		this.resultsFilepath = Constants.FILE_PATH;
		this.resultsFilename = Constants.FILENAME;
		this.debugOutputFile = Constants.DEBUG_OUPUT_FILE;
		this.debugMode = Constants.DEBUG_MODE;
		this.approximateSystem = Constants.APPROXIMATE_SYSTEM;
		this.unbindsFromIBSifFree = Constants.UNBINDS_FROM_IBS_IF_BUSY;
		this.unbindsFromterminationIfFree = Constants.UNBINDS_FROM_TERMINATION_IF_BUSY;
		this.unbindsFromCodonIfBusy=Constants.UNBINDS_FROM_CODON_IF_BUSY;
		this.stochasticmRNACopyNumber = Constants.STOCHASTIC_MRNA_COPY_NUMBER;
		this.printPolyosomeSize = false;
		this.polyosomeSizePrintPoints=Constants.POLYOSOME_SIZE_PRINT_POINTS;
		if(this.polyosomeSizePrintPoints > 0){
			printPolyosomeSize = true;
		}

		this.nonGognateIntroduceWaitingTime=Constants.NON_COGNATE_INTRODUCE_WAITING_TIME; 
		this.nonGognateWaitingTime=Constants.NON_COGNATE_WAIT_TIME;
		this.nonCognateChangeBindingRate = Constants.NON_COGNATE_CHANGE_BINDING_RATE;
		this.tRNAAbundaceFile = Constants.TRNA_ABUNDANCE_FILENAME;
		
		this.tRNAbindingRate = Constants.TRNA_BINDING_RATE;
		this.tRNAchargeRate = Constants.TRNA_CHARGE_RATE;
		this.ribosomeAffinityAvg = Constants.RIBOSOME_AFFINITY_AVG;
		this.ribosomeAffinityStd = Constants.RIBOSOME_AFFINITY_STD;
		this.IBSmoveRateMax = Constants.IBS_MOVE_RATE_MIN;
		this.IBSmoveRateMin = Constants.IBS_MOVE_RATE_MAX;
		this.noOfRibosomeSitesBefore = Constants.NO_OF_RIBOSOME_SITES_BEFORE;
		this.noOfRibosomeSitesAfter = Constants.NO_OF_RIBOSOME_SITES_AFTER;
		this.noOfRibosomeSitesTotal = Constants.NO_OF_RIBOSOME_SITES_TOTAL;
		this.noOfTranslationPerRibosomemRNAMax = Constants.NO_OF_TRANSLATION_PER_RIBOSOME_MRNA_MAX;
		this.noOfTranslationPerRibosomemRNAMin = Constants.NO_OF_TRANSLATION_PER_RIBOSOME_MRNA_MIN;
		this.elongationCognateProbability = Constants.ELONGATION_COGNATE_PROBABILITY;
		this.elongationNearCognateProbability = Constants.ELONGATION_NEARCOGNATE_PROBABILITY;
		this.elongationCognateRate = Constants.ELONGATION_COGNATE_RATE;
		this.elongationNearCognateRate = Constants.ELONGATION_NEARCOGNATE_RATE;
		this.elongationTranslocationRate = Constants.ELONGATION_TRANSLOCATION_RATE;
		this.elongationNonCognateProbability = Constants.ELONGATION_NONCOGNATE_PROBABILITY;
		this.mRNAabundanceFilename = Constants.MRNA_ABUNDANCE_FILENAME;
		this.IBSlengthAvg = Constants.IBS_LENGTH_AVG;
		this.IBSscanningRate = Constants.IBS_SCANNING_RATE;
		this.noOfCopiesPermRNAMin = Constants.NO_OF_COPIES_PER_MRNA_MIN;
		this.noOfSynthesisedProteinsPermRNAmin = Constants.NO_OF_SYNTHESISED_PROTEINS_PER_MRNA_MIN;
		this.ratioOfFreeRibosomes = Constants.RATIO_OF_FREE_RIBOSOMES;
		this.elongationBindingEventsUseGillespeDirect = Constants.ELONGATION_BINDING_EVENTS_USE_GILESPIE_DIRECT;
		this.recordStartTime = Constants.RECORD_START_TIME;
		this.elongationNearCognateProbabilityExitRecognition = Constants.ELONGATION_NEARCOGNATE_PROBABILITY_EXIT_RECOGNITION;
		this.elongationCognateProbabilityExitRecognition = Constants.ELONGATION_COGNATE_PROBABILITY_EXIT_RECOGNITION;
		this.stallEventsWait = Constants.STALL_EVENTS_WAIT;
		this.collisionsBeforeUnbinds = Constants.COLLISIONS_BEFORE_UNBINDS;
	}
	
	/**
	 * checks whether input parameters make sense
	 */
	private void checkInputDataConsistency(){
		if(this.unbindsFromCodonIfBusy || this.unbindsFromIBSifFree || this.unbindsFromterminationIfFree){
			this.stallEventsWait = false;
		}
		if(this.stallEventsWait){
			this.printDebugInfo("Collisions are not scheduled");
		} else{
			this.printDebugInfo("Collisions are possible");

		}
	}
	
	/**
	 * prints to the debug file or screen the debug message
	 * @param str the string to print
	 */
	public void printDebugInfo(String str){
		printDebugInfo(str, true);
	}
	
	public void printDebugInfo(String str, boolean append){
		if(!this.debugOutputFile.isEmpty()){
			try {
				Utils.writeToFile(this.resultsFilepath+this.debugOutputFile, str, append);
			} catch (Exception e) {
				System.err.println(e);
			}
		} else{
			System.out.println(str);
		}
	}
	

	private void printRates(){
		printDebugInfo("stopTime = "+this.stopTime);
		printDebugInfo("nonGognateWaitingTime = "+this.nonGognateWaitingTime);
		printDebugInfo("tRNAbindingRate = "+this.tRNAbindingRate);
		printDebugInfo("tRNAchargeRate = "+this.tRNAchargeRate);
		printDebugInfo("ribosomeAffinityAvg = "+this.ribosomeAffinityAvg);
		printDebugInfo("ribosomeAffinityStd = "+this.ribosomeAffinityStd);
		printDebugInfo("IBSmoveRateMin = "+this.IBSmoveRateMin);
		printDebugInfo("IBSmoveRateMax = "+this.IBSmoveRateMax);
		printDebugInfo("elongationCognateRate = "+this.elongationCognateRate);
		printDebugInfo("elongationNearCognateRate = "+this.elongationNearCognateRate);
		printDebugInfo("elongationTranslocationRate = "+this.elongationTranslocationRate);
		printDebugInfo("====================================");
	}
	
	
	/**
	 * prints output information for each codons
	 * @throws FileNotFoundException 
	 */
	private void printCodonInformation() throws FileNotFoundException{
		countCodons();
		String str,filename;
		//filename = this.resultsFilepath+this.resultsFilename+"_codon.txt";
		filename = this.resultsFilepath+ this.parameterFile.getName().replaceAll("params", "results_codon").replaceAll("txt", "csv");
		printDebugInfo("codon information printed to file: "+filename,true);
		str="#codonID, codon, count, frequency, cognateElongated, nearCognateElongated, totalElongated, cognateElongationTimeTotal, nearCognateElongationTimeTotal, ElongationTimeTotal, cognateElongationTimeAvg, nearCognateElongationTimeAvg, ElongationTimeAvg, cpognateAttemptsTotal, nearCognateAttemptsTotal, cpognateAttempts, nearCognateAttempts";
		Utils.writeToFile(filename,str,false);
		//printDebugInfo();
		for(int i=0;i<SequenceUtils.codons.codon.length;i++){
			str = i+"";
			str +=", "+SequenceUtils.codons.codon[i];
			str +=", "+this.codonCount[i];
			str +=", "+((double)this.codonCount[i]/this.codonCountTotal);
			str +=", "+this.cognateElongated[i];
			str +=", "+this.nearCognateElongated[i];
			str +=", "+(this.cognateElongated[i]+this.nearCognateElongated[i]);
			str +=", "+this.cognateElongationTimeTotal[i];
			str +=", "+this.nearCognateElongationTimeTotal[i];
			str +=", "+(this.cognateElongationTimeTotal[i]+this.nearCognateElongationTimeTotal[i]);
			str +=", "+((double)this.cognateElongationTimeTotal[i]/this.cognateElongated[i]);
			str +=", "+((double)this.nearCognateElongationTimeTotal[i]/this.nearCognateElongated[i]);
			str +=", "+((double)(this.cognateElongationTimeTotal[i]+this.nearCognateElongationTimeTotal[i])/(this.cognateElongated[i]+this.nearCognateElongated[i]));
			str +=", "+this.cognateAttemptsTotal[i];
			str +=", "+this.nearCognateAttemptsTotal[i];
			str +=", "+((double)this.cognateAttemptsTotal[i]/(this.cognateElongated[i]+this.nearCognateElongated[i]));
			str +=", "+((double)this.nearCognateAttemptsTotal[i]/(this.cognateElongated[i]+this.nearCognateElongated[i]));
			Utils.writeToFile(filename,str,true);
		}
	}
	/**
	 * count codons
	 */
	private void countCodons(){
		this.codonCountTotal=0;
		for(int i=0;i<this.mRNAtype.length;i++){
			for(int j=0;j< this.mRNAtype[i].getLength();j++){
				this.codonCount[this.mRNAtype[i].getCodon(j)]+=this.mRNAtype[i].getCopyNumber();
				this.codonCountTotal+=this.mRNAtype[i].getCopyNumber();
			}
		}
		
	}
	
	
	
	/**
	 * prints output information for each mRNA
	 * @throws FileNotFoundException 
	 */
	private void printMRNAInformation() throws FileNotFoundException{
		String str,filename;
		//countRibosomesPerMRNA();
		//filename = this.resultsFilepath+this.resultsFilename+"_mRNA.txt";
		filename = this.resultsFilepath+ this.parameterFile.getName().replaceAll("params", "results_mRNA").replaceAll("txt", "csv");
		printDebugInfo("mRNA information printed to file: "+filename,true);
		str="#mRNAtype, name, translated, productionRate(molecules/min), ribosomeAffinity, IBSlength, IBSmoverate, length, copyNumber, polyosomeSize, polyosomeSizeAvg, polyosomeSizeCount, totalNumberOfPositions, relativePolyosomeSize";
		Utils.writeToFile(filename,str,false);

		for(int i=0;i<mRNATranslated.length;i++){
			str=i+"";
			str+=", "+mRNAtype[i].getName();
			str+=", "+mRNATranslated[i];
			str+=", "+((double)(mRNATranslated[i]*60)/((this.stopTime-this.recordStartTime)));
			str+=", "+this.mRNAtype[i].getRibosomeAffinity();
			str+=", "+this.mRNAtype[i].getIBSlength();
			str+=", "+this.mRNAtype[i].getIBSmoveRate();
			str+=", "+this.mRNAtype[i].getLength();
			str+=", "+this.mRNAtype[i].getCopyNumber();		
			str+=", "+this.ribosomesPerMRNA[i];	
			str+=", "+this.ribosomesPerMRNAAvg[i];	
			str+=", "+this.ribosomesPerMRNACount[i];	
			str+=", "+SequenceUtils.getTotalNoOfPositions(this.mRNAtype[i].getLength(), this.noOfRibosomeSitesAfter,this.noOfRibosomeSitesBefore);
			str+=", "+SequenceUtils.getRelativeRibosomeOccupancy(this.ribosomesPerMRNAAvg[i],this.mRNAtype[i].getLength(),this.mRNAtype[i].getCopyNumber(), this.noOfRibosomeSitesAfter,this.noOfRibosomeSitesBefore);	
			Utils.writeToFile(filename,str,true);
		}
	}
	

	
	/**
	 * counts the mRNA ribosome occupancy
	 */
	private void countRibosomesPerMRNA(){
		for(int i=0;i<this.ribosomes.length;i++){
			if(this.ribosomes[i].getmRNA()>=0){
				this.ribosomesPerMRNA[this.mRNA[this.ribosomes[i].getmRNA()].getType()]++;
			}
		}
	}
	
	/**
	 * prints output information for each tRNA
	 * @throws FileNotFoundException 
	 */
	private void printTRNAInformation() throws FileNotFoundException{
		String str,filename;
		countRibosomesPerMRNA();
		int totalTRNA = this.tRNATabe.getTotalAbundance();
		int chargedTRNA = this.tRNATabe.getChargedAbundance();
		
		filename = this.resultsFilepath+ this.parameterFile.getName().replaceAll("params", "results_tRNA").replaceAll("txt", "csv");
		printDebugInfo("tRNA information printed to file: "+filename,true);

		//filename = this.resultsFilepath+this.resultsFilename+"_tRNA.txt";
		str="#tRNAID, tRNA, total, relativeTotal, chargedCount, chargedSS,  busyCount, busySS, unchargedCount, unchargedSS";
		int abundance;
		Utils.writeToFile(filename,str,false);
		for(int i=0;i<SequenceUtils.tRNAtable.getNoOftRNAs();i++){
			abundance=this.tRNATabe.gettRNAChargeAbundance(i)+this.tRNATabe.gettRNABusyAbundance(i)+this.tRNATabe.gettRNAunchargeAbundance(i);
			if(abundance != 0){
				str=""+i;
				str+=", t"+i;
				str+=", "+abundance;
				//str+=", "+this.tRNATabe.gettRNAabundance(i);
				str+=", "+((double)abundance/totalTRNA);
		//		str+=", "+(this.tRNAchargedAvg[i]*this.tRNAchargedCount[i]);
				str+=", "+this.tRNAchargedCount[i];
				str+=", "+this.tRNAchargedAvg[i];
		//		str+=", "+(this.tRNAbusyAvg[i]*this.tRNAbusyCount[i]);
				str+=", "+this.tRNAbusyCount[i];
				str+=", "+this.tRNAbusyAvg[i];
		//		str+=", "+(this.tRNAunchargedAvg[i]*this.tRNAunchargedCount[i]);
				str+=", "+this.tRNAunchargedCount[i];
				str+=", "+ this.tRNAunchargedAvg[i];
				Utils.writeToFile(filename,str,true);
			}
			/*if(this.tRNAchargedTotal[i]<0){
				this.printDebugInfo("err: t"+i);
			}*/
		}
	}
	
	/**
	 * copys the parameter file to a new file
	 * @param parameterFilename
	 */
	private void generateParameterFile(String parameterFilename){

		    try {
		    	File directory = new File(this.resultsFilepath);
		        // Create temporary file.
		        this.parameterFile = File.createTempFile("params_", ".txt",directory);
		        
		        //System.out.println(parameterFile.getName() );
		        this.debugOutputFile = this.parameterFile.getName().replaceAll("params", "status");
		        
		        // Write to temp file
		        BufferedWriter out = new BufferedWriter(new FileWriter(parameterFile));
		        out.write("stopTime="+this.stopTime+";\n");
		        out.write("noOfmRNAtype="+this.noOfmRNAtype+";\n");
		        out.write("noOfCopiesPermRNAAvg="+this.noOfCopiesPermRNAAvg+";\n");
		        out.write("mRNAcodonLengthMin="+this.mRNAcodonLengthMin+";\n");
		        out.write("mRNAcodonLengthMax="+this.mRNAcodonLengthMax+";\n");
		        out.write("randommRNASeed="+this.randommRNASeed+";\n");
		        out.write("mRNAfilename="+this.mRNAfilename+";\n");
		        out.write("noOfRibosomes="+this.noOfRibosomes+";\n");
		        out.write("resultsFilepath="+this.resultsFilepath+";\n");
		        out.write("resultsFilename="+this.resultsFilename+";\n");
		        out.write("debugOutputFile="+this.debugOutputFile+";\n");
		        out.write("debugMode="+this.debugMode+";\n");
		        out.write("approximateSystem="+this.approximateSystem+";\n");
		        out.write("unbindsFromIBSifFree="+this.unbindsFromIBSifFree+";\n");
		        out.write("unbindsFromterminationIfFree="+this.unbindsFromterminationIfFree+";\n");
		        out.write("unbindsFromCodonIfBusy="+this.unbindsFromCodonIfBusy+";\n");
		        out.write("stochasticmRNACopyNumber="+this.stochasticmRNACopyNumber+";\n");
		        out.write("printPolyosomeSize="+this.printPolyosomeSize+";\n");
		        out.write("polyosomeSizePrintPoints="+this.polyosomeSizePrintPoints+";\n");
		        out.write("printPolyosomeSize="+this.printPolyosomeSize+";\n");
		        out.write("nonGognateIntroduceWaitingTime="+this.nonGognateIntroduceWaitingTime+";\n");
		        out.write("nonGognateWaitingTime="+this.nonGognateWaitingTime+";\n");
		        out.write("nonCognateChangeBindingRate="+this.nonCognateChangeBindingRate+";\n");
		        out.write("tRNAAbundaceFile="+this.tRNAAbundaceFile+";\n");
		        out.write("tRNAbindingRate="+this.tRNAbindingRate+";\n");
		        out.write("tRNAchargeRate="+this.tRNAchargeRate+";\n");
		        out.write("ribosomeAffinityAvg="+this.ribosomeAffinityAvg+";\n");
		        out.write("ribosomeAffinityStd="+this.ribosomeAffinityStd+";\n");
		        out.write("IBSmoveRateMax="+this.IBSmoveRateMax+";\n");
		        out.write("IBSmoveRateMin="+this.IBSmoveRateMin+";\n");
		        out.write("noOfRibosomeSitesBefore="+this.noOfRibosomeSitesBefore+";\n");
		        out.write("noOfRibosomeSitesAfter="+this.noOfRibosomeSitesAfter+";\n");
		        out.write("noOfRibosomeSitesTotal="+this.noOfRibosomeSitesTotal+";\n");
		        out.write("noOfTranslationPerRibosomemRNAMax="+this.noOfTranslationPerRibosomemRNAMax+";\n");
		        out.write("noOfTranslationPerRibosomemRNAMin="+this.noOfTranslationPerRibosomemRNAMin+";\n");
		        out.write("elongationCognateProbability="+this.elongationCognateProbability+";\n");
		        out.write("elongationNearCognateProbability="+this.elongationNearCognateProbability+";\n");
		        out.write("elongationCognateRate="+this.elongationCognateRate+";\n");
		        out.write("elongationNearCognateRate="+this.elongationNearCognateRate+";\n");
		        out.write("elongationTranslocationRate="+this.elongationTranslocationRate+";\n");
		        out.write("elongationNonCognateProbability="+this.elongationNonCognateProbability+";\n");
		        out.write("mRNAabundanceFilename="+this.mRNAabundanceFilename+";\n");
		        out.write("IBSlengthAvg="+this.IBSlengthAvg+";\n");
		        out.write("IBSscanningRate="+this.IBSscanningRate+";\n");
		        out.write("noOfCopiesPermRNAMin="+this.noOfCopiesPermRNAMin+";\n");
		        out.write("noOfSynthesisedProteinsPermRNAmin="+this.noOfSynthesisedProteinsPermRNAmin+";\n");
		        out.write("ratioOfFreeRibosomes="+this.ratioOfFreeRibosomes+";\n");
		        out.write("elongationBindingEventsUseGillespeDirect="+this.elongationBindingEventsUseGillespeDirect+";\n");
		        out.write("recordStartTime="+this.recordStartTime+";\n");
		        out.write("elongationNearCognateProbabilityExitRecognition="+this.elongationNearCognateProbabilityExitRecognition+";\n");
		        out.write("elongationCognateProbabilityExitRecognition="+this.elongationCognateProbabilityExitRecognition+";\n");
		        out.write("stallEventsWait="+this.stallEventsWait+";\n");
		        out.write("collisionsBeforeUnbinds="+this.collisionsBeforeUnbinds+";\n");
		
		        out.close();

		    } catch (IOException e) {
		    	e.printStackTrace();
		    }
	}
	
}
