package objects;

import java.util.Random;
import utils.SequenceUtils;

/**
 * A class in which we store the data on mRNA and 
 * @author nrz2
 *
 */

public class MessengerRNAType {

	private String name; //the name of the mRNA 
	private byte[] strand; //the mRNA strand defined by characters
	private double ribosomeAffinity; // a constant that determines the affinity of the ribosomes for this particular strand
	private int copyNumber;

	private double IBSmoveRate;

	private double IBSlength;
	private double ssAbundance;
	private double proteinSynthesisRate;
	
	/**
	 * constructs a random mRNA sequence
	 * @param name the name of the sequence;
	 */
	public MessengerRNAType(String name, Random generator, int mRNAcodonLengthMin, int mRNAcodonLengthMax, double ribosomeAffinityAvg, double IBSmoveRateMin, double IBSmoveRateMax){
		this.name = name;

		int length= mRNAcodonLengthMin+generator.nextInt(mRNAcodonLengthMax - mRNAcodonLengthMin);
		//System.out.println(length);
		strand = SequenceUtils.randomMessengerRNA(length, generator);
		
		
		/*ribosomeAffinity = (double)((int)((Constants.RIBOSOME_AFFINITY_AVG + generator.nextGaussian()*Constants.RIBOSOME_AFFINITY_STD)*100))/100;
		if(ribosomeAffinity<0){
			ribosomeAffinity = 0;
		}*/
		IBSmoveRate =  IBSmoveRateMin+generator.nextDouble()*(IBSmoveRateMax - IBSmoveRateMin);
		
		ribosomeAffinity = ribosomeAffinityAvg;
		//IBSmoveRate = Constants.IBS_MOVE_RATE_MAX + Constants.IBS_MOVE_RATE_MIN/2;*/
		this.copyNumber=1;
		IBSlength = -1;
		ssAbundance = -1;
		proteinSynthesisRate = -1;
	}
	
	/**
	 * constructs a new mRNA type from a specified name and strand
	 * @param name the name of the mRNA
	 * @param strand the byte array with the codon sequence
	 */
	public MessengerRNAType(String name, byte[] strand){
		this.name = name;

				
		this.strand =new byte[strand.length];
		for(int i=0;i<strand.length;i++){
			this.strand[i] = strand[i];
		}
		IBSmoveRate =  0;
		ribosomeAffinity = 0;
		this.copyNumber=1;
		IBSlength = -1;
		ssAbundance = -1;
		proteinSynthesisRate = -1;
	}
	
	
	/**
	 * contructor that stores the name and various other parameters.
	 * @param name
	 * @param IBSlength
	 * @param ssAbundance
	 * @param proteinSynthesisRate
	 */
	public MessengerRNAType(String name, double IBSlength, double ssAbundance, double proteinSynthesisRate){
		this.name = name;
		this.strand =null;
		IBSmoveRate =  0;
		ribosomeAffinity = 0;
		this.copyNumber=1;
		this.IBSlength = IBSlength;
		this.ssAbundance = ssAbundance;
		this.proteinSynthesisRate = proteinSynthesisRate;
	}
	
	/**
	 * sets the value of the IBS region
	 * @param IBSlength
	 */
	public void setIBSlength(double IBSlength){
            //IBSLength is the first col in biodata/5UTRxxx.csv
		this.IBSlength=IBSlength;//strictly obsolete now.
                this.IBSmoveRate = IBSlength; //change by CHU
	}
	
	/**
	 * returns the length of the IBS region
	 * @return
	 */
	public double getIBSlength(){
		return IBSlength;
	}
	
	/**
	 * sets the steady state abundance of the mRNA
	 * @param IBSlength
	 */
	public void setSSabundace(double ssAbundance){
		this.ssAbundance=ssAbundance;
	}
	
	/**
	 * returns the length of the IBS region
	 * @return
	 */
	public double getSSabundace(){
		return ssAbundance;
	}
	
	/**
	 * sets the steady state protein synthesis rate of the mRNA
	 * @param IBSlength
	 */
	public void setProteinSynthesisRate(double proteinSynthesisRate){
		this.proteinSynthesisRate=proteinSynthesisRate;
	}
	
	/**
	 * returns the steady state protein synthesis rate of the mRNA
	 * @return
	 */
	public double getProteinSynthesisRate(){
		return proteinSynthesisRate;
	}

	/**
	 * sets the copy number of this mRNA species
	 * @param copyNumber
	 */
	public void setCopyNumber(int copyNumber){
		this.copyNumber = copyNumber;
	}
	
	public int getCopyNumber(){
		return this.copyNumber;
	}
	
	/**
	 * constructor that initialises the mRNA with user provided data
	 * @param name the name of the mRNA strand
	 * @param strand the strand sequence
	 * @param ribosomeAffinity the ribosome affinity
	 * @param IBSmoveRate the rate at which the ribosome moves from the IBS
	 */
	public MessengerRNAType(String name,byte[] strand,double ribosomeAffinity, double IBSmoveRate){
		this.name = name;

		this.strand = new byte[strand.length];
		for(int i=0;i<strand.length;i++){
			this.strand[i]=strand[i];
		}                      
		this.ribosomeAffinity = ribosomeAffinity;
		this.IBSmoveRate = IBSmoveRate;
	}
	
	/**
	 * accessor method that return the ribosome affinity for this particular mRNA strand
	 * @return ribosome affinity 
	 */
	public double getRibosomeAffinity(){
		return ribosomeAffinity;
	}
	
	
	/**
	 * mutator method that sets a new value for the ribosome affinity for this particular mRNA strand
	 * @param ribosomeAffinity the new value for the ribosome affinity
	 */
	public void setRibosomeAffinity(double ribosomeAffinity){
		this.ribosomeAffinity = ribosomeAffinity;
                //System.out.println(ribosomeAffinity);
	}
	
	/**
	 * accessor method to return the name of the mRNA
	 * @return the name of the mRNA
	 */
	public String getName(){
		return name;
	}
	
	
	/**
	 * mutator method that changes the name of the strand
	 * @param name the new name for the strand
	 */
	public void setName(String name){
		this.name = name;
	}
	
	/**
	 * gets the codon at a specific position 
	 * @param position the position on the strand
	 * @return the codon found at the specified position
	 */
	public byte getCodon(int position){
		if(position>=0 && position< strand.length){
			return strand[position];
		} else{
			return -2;
		}
	}
	
	/**
	 * method that returns the number of strands in the sequence
	 * @return number of strands in the sequence
	 */
	public int getLength(){
		return strand.length;
	}
	
	/**
	 * method that sets a new value for the rate to leave the IBS site
	 * @param IBSmoveRate the new rate
	 */
	public void setIBSmoveRate(double IBSmoveRate){
           
		this.IBSmoveRate = IBSmoveRate;
                this.IBSmoveRate = getIBSlength() ;
                //CHU: In essence, the IBSMoveRate is the length. the length has lost any meaning. We no longer need it. 
                //The original idea by Radu was that we use the length to calculate the
                //Initiation time. however, this is not necessary now because
                //We assume the initiation time to be know. 
	 //System.out.println(this.IBSmoveRate + "<----");
        }
	
	/**
	 * method that returns the rate to leave the IBS site;
	 * @return the rate to leave the IBS site;
	 */
	public double getIBSmoveRate(){
                //As above. The length is now the rate. 
	    return getIBSlength();
                //for explanation see: Cell.java :  loadmRNADataFromFile
                // and 
                
	}
}
